import random
import dspy
from rich.progress import Progress
from typing import AsyncGenerator

# Weaviate
from weaviate.classes.aggregate import GroupByAggregate
from weaviate.classes.query import Metrics, Filter
from weaviate.collections import CollectionAsync
from weaviate.classes.config import Configure, Property, DataType, Tokenization

from elysia.config import nlp, Settings, load_base_lm, ElysiaKeyManager
from elysia.config import settings as environment_settings

from elysia.util import return_types as rt

from elysia.preprocessing.prompt_templates import (
    CollectionSummariserPrompt,
    DataMappingPrompt,
    ReturnTypePrompt,
    PromptSuggestorPrompt,
)
from elysia.util.collection import async_get_collection_data_types
from elysia.util.async_util import asyncio_run
from elysia.util.parsing import format_dict_to_serialisable
from elysia.util.client import ClientManager


class ProcessUpdate:
    def __init__(self, collection_name: str, total: int) -> None:
        self.collection_name = collection_name
        self.total = total
        self.progress = 0

    def update_progress(self) -> None:
        self.progress += 1

    def update_total(self, total: int) -> None:
        self.total = total

    async def to_frontend(
        self, completed: bool = False, message: str = "", error: str = "", **kwargs
    ) -> dict:
        return {
            "type": ("update" if not completed else "completed"),
            "collection_name": self.collection_name,
            "progress": min(self.progress / self.total, 0.99) if not completed else 1.0,
            "message": message,
            "error": error,
        }

    async def __call__(self, *args, **kwargs) -> dict:
        self.update_progress()
        return await self.to_frontend(*args, **kwargs)


async def _summarise_collection(
    collection_summariser_prompt: dspy.Module,
    properties: dict,
    subset_objects: list[dict],
    len_collection: int,
    settings: Settings,
    lm: dspy.LM,
) -> tuple[str, dict]:

    with ElysiaKeyManager(settings):
        prediction = await collection_summariser_prompt.aforward(
            data_sample=subset_objects,
            data_fields=list(properties.keys()),
            sample_size=f"""
            {len(subset_objects)} out of a total of {len_collection}.
            You are seeing {round(len(subset_objects)/len_collection*100, 2)}% of the objects.
            """,
            lm=lm,
        )

    summary_concat = ""
    for sentence in [
        prediction.overall_summary,
        prediction.relationships,
        prediction.structure,
        prediction.irregularities,
    ]:
        if (
            sentence.endswith(".")
            or sentence.endswith("?")
            or sentence.endswith("!")
            or sentence.endswith("\n")
        ):
            summary_concat += f"{sentence} "
        else:
            summary_concat += f"{sentence}."

    return summary_concat, prediction.field_descriptions


async def _evaluate_field_statistics(
    collection: CollectionAsync,
    properties: dict,
    property: str,
    len_collection: int,
    sample_objects: list[dict],
) -> dict:
    out = {}
    out["type"] = properties[property] if properties[property] != "number" else "float"
    out["name"] = property
    out["range"] = None
    out["mean"] = None
    out["date_range"] = None
    out["date_median"] = None
    out["groups"] = None

    if not out["type"].startswith("object"):

        groups_response = await collection.aggregate.over_all(
            total_count=True, group_by=GroupByAggregate(prop=property, limit=30)
        )

        if len(groups_response.groups) > 15:
            groups = [
                {
                    "value": str(
                        group.grouped_by.value
                    ),  # must force to string for weaviate
                    "count": group.total_count,
                }
                for group in groups_response.groups
                if group.total_count > 1
            ]
        else:
            groups = [
                {
                    "value": str(group.grouped_by.value),
                    "count": group.total_count,
                }
                for group in groups_response.groups
            ]

        total_count = sum(group["count"] for group in groups)
        remainder = len_collection - total_count
        group_coverage = total_count / len_collection
        if remainder > 0:
            groups.append(
                {
                    "value": "<undefined_group> (not used for filtering)",
                    "count": remainder,
                }
            )

        # check if groups are useful
        if len(groups) == 1 or group_coverage < 0.5:
            out["groups"] = None
        else:
            out["groups"] = groups

    # Number (summary statistics)
    if properties[property] == "int":
        response = await collection.aggregate.over_all(
            return_metrics=[
                Metrics(property).integer(mean=True, maximum=True, minimum=True),
            ]
        )
        out["range"] = [
            response.properties[property].minimum,
            response.properties[property].maximum,
        ]
        out["mean"] = response.properties[property].mean
        out["date_range"] = None
        out["date_median"] = None

    elif properties[property] == "number":
        response = await collection.aggregate.over_all(
            return_metrics=[
                Metrics(property).number(mean=True, maximum=True, minimum=True),
            ]
        )
        out["range"] = [
            response.properties[property].minimum,
            response.properties[property].maximum,
        ]
        out["mean"] = response.properties[property].mean
        out["date_range"] = None
        out["date_median"] = None

    # Text (lengths)
    elif properties[property] == "text":

        # For text, we want to evaluate the length of the text in tokens (use spacy)
        lengths = []
        for obj in sample_objects:
            if property in obj and isinstance(obj[property], str):
                lengths.append(len(nlp(obj[property])))

        if len(lengths) == 0:
            out["range"] = None
            out["mean"] = None
        else:
            out["range"] = [min(lengths), max(lengths)]
            out["mean"] = sum(lengths) / len(lengths)
        out["date_median"] = None
        out["date_range"] = None

    # Boolean (grouping + mean)
    elif properties[property] == "boolean":
        response = await collection.aggregate.over_all(
            return_metrics=[Metrics(property).boolean(percentage_true=True)]
        )
        out["mean"] = response.properties[property].percentage_true
        out["range"] = [0, 1]
        out["date_range"] = None
        out["date_median"] = None

    # Date (summary statistics)
    elif properties[property] == "date":
        response = await collection.aggregate.over_all(
            return_metrics=[
                Metrics(property).date_(median=True, minimum=True, maximum=True)
            ]
        )
        out["date_range"] = [
            response.properties[property].minimum,
            response.properties[property].maximum,
        ]
        out["mean"] = None
        out["date_median"] = response.properties[property].median
        out["range"] = None
        out["mean"] = None

    # List (lengths)
    elif properties[property].endswith("[]"):
        lengths = [len(obj[property]) for obj in sample_objects]

        if len(lengths) == 0:
            out["range"] = None
            out["mean"] = None
        else:
            out["range"] = [min(lengths), max(lengths)]
            out["mean"] = sum(lengths) / len(lengths)
        out["date_range"] = None
        out["date_median"] = None

    return out


async def _suggest_prompts(
    prompt_suggestor_prompt: dspy.Module,
    collection_information: dict,
    example_objects: list[dict],
    settings: Settings,
    lm: dspy.LM,
) -> list[str]:
    with ElysiaKeyManager(settings):
        prediction = await prompt_suggestor_prompt.aforward(
            collection_information=collection_information,
            example_objects=example_objects,
            lm=lm,
        )
    return prediction.prompt_suggestions


async def _evaluate_return_types(
    return_type_prompt: dspy.Module,
    collection_summary: str,
    data_fields: dict,
    example_objects: list[dict],
    settings: Settings,
    lm: dspy.LM,
) -> list[str]:

    with ElysiaKeyManager(settings):
        prediction = await return_type_prompt.aforward(
            collection_summary=collection_summary,
            data_fields=data_fields,
            example_objects=example_objects,
            possible_return_types=rt.specific_return_types,
            lm=lm,
        )
    return_types = prediction.return_types

    if return_types == []:
        return_types = ["generic"]

    # if conversation is selected, then message should also be selected
    if "conversation" in return_types and "message" not in return_types:
        return_types.append("message")

    return return_types


async def _define_mappings(
    data_mapping_prompt: dspy.Module,
    input_fields: list,
    output_fields: list,
    properties: dict,
    collection_information: dict,
    example_objects: list[dict],
    mapping_type: str,
    settings: Settings,
    lm: dspy.LM,
) -> dict:

    with ElysiaKeyManager(settings):
        prediction = await data_mapping_prompt.aforward(
            mapping_type=mapping_type,
            input_data_fields=input_fields,
            output_data_fields=output_fields,
            input_data_types=properties,
            collection_information=collection_information,
            example_objects=example_objects,
            lm=lm,
        )

    return prediction.field_mapping


async def _evaluate_index_properties(collection: CollectionAsync) -> dict:
    schema_info = await collection.config.get()

    index_properties = {
        "isNullIndexed": schema_info.inverted_index_config.index_null_state,
        "isLengthIndexed": schema_info.inverted_index_config.index_property_length,
        "isTimestampIndexed": schema_info.inverted_index_config.index_timestamps,
    }
    return index_properties


async def _find_vectorisers(collection: CollectionAsync) -> dict[str, dict]:
    schema_info = await collection.config.get()
    if not schema_info.vector_config:
        named_vectors = None
    else:
        named_vectors = [
            {
                "name": vector,
                "vectorizer": schema_info.vector_config[
                    vector
                ].vectorizer.vectorizer.name,
                "model": (
                    schema_info.vector_config[vector].vectorizer.model["model"]
                    if "model" in schema_info.vector_config[vector].vectorizer.model
                    else None
                ),
                "source_properties": schema_info.vector_config[
                    vector
                ].vectorizer.source_properties,
                "enabled": True,
                "description": "",
            }
            for vector in schema_info.vector_config
        ]

    if not schema_info.vectorizer_config:
        vectoriser = None
    else:
        vectoriser = {
            "vectorizer": schema_info.vectorizer_config.vectorizer.name,
            "model": (
                schema_info.vectorizer_config.model["model"]
                if "model" in schema_info.vectorizer_config.model
                else None
            ),
        }

    return named_vectors, vectoriser


async def preprocess_async(
    collection_name: str,
    client_manager: ClientManager | None = None,
    min_sample_size: int = 10,
    max_sample_size: int = 20,
    num_sample_tokens: int = 30000,
    force: bool = False,
    percentage_correct_threshold: float = 0.3,
    settings: Settings = environment_settings,
) -> AsyncGenerator[dict, None]:
    """
    Preprocess a collection, obtain a LLM-generated summary of the collection,
    a set of statistics for each field (such as unique categories), and a set of mappings
    from the fields to the frontend-specific fields in Elysia.

    In order:

    1. Evaluate all the data fields and groups/statistics of the data fields as a whole
    2. Write a summary of the collection via an LLM
    3. Evaluate what return types are available for this collection
    4. For each data field in the collection, evaluate what corresponding entry goes to what field in the return type (mapping)
    5. Save as a ELYSIA_METADATA__ collection

    Args:
        collection_name (str): The name of the collection to preprocess.
        client_manager (ClientManager): The client manager to use.
        min_sample_size (int): The minimum number of objects to sample from the collection to evaluate the statistics/summary. Optional, defaults to 10.
        max_sample_size (int): The maximum number of objects to sample from the collection to evaluate the statistics/summary. Optional, defaults to 20.
        num_sample_tokens (int): The number of tokens to approximately sample from the collection to evaluate the summary.
            The preprocessor will aim to use this many tokens in the sample objects to evaluate the summary.
            But will not exceed the maximum number of objects specified by `max_sample_size`, and always use at least `min_sample_size` objects.
        force (bool): Whether to force the preprocessor to run even if the collection already exists. Optional, defaults to False.
        threshold_for_missing_fields (float): The threshold for the number of missing fields in the data mapping. Optional, defaults to 0.1.
        settings (Settings): The settings to use. Optional, defaults to the environment variables/configured settings.

    Returns:
        AsyncGenerator[dict, None]: A generator that yields dictionaries with the status updates and progress of the preprocessor.
    """

    collection_summariser_prompt = dspy.ChainOfThought(CollectionSummariserPrompt)
    return_type_prompt = dspy.ChainOfThought(ReturnTypePrompt)
    data_mapping_prompt = dspy.ChainOfThought(DataMappingPrompt)
    prompt_suggestor_prompt = dspy.ChainOfThought(PromptSuggestorPrompt)

    lm = load_base_lm(settings)
    logger = settings.logger
    process_update = ProcessUpdate(collection_name, len(rt.specific_return_types) + 5)

    if client_manager is None:
        client_manager = ClientManager(
            wcd_url=settings.WCD_URL, wcd_api_key=settings.WCD_API_KEY
        )
        close_clients_after_completion = True
    else:
        close_clients_after_completion = False

    try:
        # Check if the collection exists
        async with client_manager.connect_to_async_client() as client:
            if not await client.collections.exists(collection_name):
                raise Exception(f"Collection {collection_name} does not exist!")

        # Check if the preprocessed collection exists
        if (
            await preprocessed_collection_exists_async(collection_name, client_manager)
            and not force
        ):
            logger.info(f"Preprocessed metadata for {collection_name} already exists!")
            return

        # Get the collection and its properties
        async with client_manager.connect_to_async_client() as client:
            collection = client.collections.get(collection_name)
            properties = await async_get_collection_data_types(client, collection_name)

        # get number of items in collection
        agg = await collection.aggregate.over_all(total_count=True)
        len_collection: int = agg.total_count  # type: ignore

        # Randomly sample sample_size objects for the summary
        indices = random.sample(
            range(len_collection),
            max(min(max_sample_size, len_collection), 1),
        )

        # Get first object to estimate token count
        obj = await collection.query.fetch_objects(limit=1, offset=indices[0])
        token_count_0 = len(nlp(str(obj.objects[0].properties)))
        subset_objects: list[dict] = [obj.objects[0].properties]  # type: ignore

        # Get number of objects to sample to get close to num_sample_tokens
        num_sample_objects = max(min_sample_size, num_sample_tokens // token_count_0)

        for index in indices[1:num_sample_objects]:
            obj = await collection.query.fetch_objects(limit=1, offset=index)
            subset_objects.append(obj.objects[0].properties)  # type: ignore

        # Estimate number of tokens
        logger.debug(
            f"Estimated token count of sample: {token_count_0*len(subset_objects)}"
        )
        logger.debug(f"Number of objects in sample: {len(subset_objects)}")

        # Summarise the collection using LLM and the subset of the data
        summary, field_descriptions = await _summarise_collection(
            collection_summariser_prompt,
            properties,
            subset_objects,
            len_collection,
            settings,
            lm,
        )

        yield await process_update(
            message="Generated summary of collection",
        )

        if len_collection > max_sample_size:
            full_response = subset_objects
        else:
            weaviate_resp = await collection.query.fetch_objects(limit=len_collection)
            full_response = [obj.properties for obj in weaviate_resp.objects]

        # Initialise the output
        named_vectors, vectoriser = await _find_vectorisers(collection)
        out = {
            "name": collection_name,
            "length": len_collection,
            "summary": summary,
            "index_properties": await _evaluate_index_properties(collection),
            "named_vectors": named_vectors,
            "vectorizer": vectoriser,
            "fields": [],
            "mappings": {},
        }

        # Evaluate the summary statistics of each field
        for property in properties:
            out["fields"].append(
                await _evaluate_field_statistics(
                    collection, properties, property, len_collection, full_response
                )
            )
            if property in field_descriptions:
                out["fields"][-1]["description"] = field_descriptions[property]
            else:
                out["fields"][-1]["description"] = ""

        yield await process_update(
            message="Evaluated field statistics",
        )

        return_types = await _evaluate_return_types(
            return_type_prompt,
            summary,
            properties,
            subset_objects,
            settings,
            lm,
        )

        yield await process_update(
            message="Evaluated return types",
        )
        process_update.update_total(len(return_types) + 5)

        # suggest prompts
        out["prompts"] = await _suggest_prompts(
            prompt_suggestor_prompt,
            out,
            subset_objects,
            settings,
            lm,
        )

        yield await process_update(
            message="Created suggestions for prompts",
        )

        # For each return type created above, define the mappings from the properties to the frontend types
        mappings = {}
        for return_type in return_types:
            fields = rt.types_dict[return_type]

            mapping = await _define_mappings(
                data_mapping_prompt,
                mapping_type=return_type,
                input_fields=list(fields.keys()),
                output_fields=list(properties.keys()),
                properties=properties,
                collection_information=out,
                example_objects=subset_objects,
                settings=settings,
                lm=lm,
            )

            # remove any extra fields the model may have added
            mapping = {k: v for k, v in mapping.items() if k in list(fields.keys())}

            yield await process_update(
                message=f"Defined mappings for {return_type}",
            )

            mappings[return_type] = mapping

        new_return_types = []
        for return_type in return_types:

            # check if the `conversation_id` field is in the mapping (required for conversation type)
            if return_type == "conversation" and (
                (
                    mappings[return_type]["conversation_id"] is None
                    or mappings[return_type]["conversation_id"] == ""
                )
                or (
                    mappings[return_type]["message_id"] is None
                    or mappings[return_type]["message_id"] == ""
                )
            ):
                continue

            # If less than threshold_for_missing_fields%, keep the return type
            num_missing = sum([m == "" for m in list(mappings[return_type].values())])
            perc_correct = 1 - (num_missing / len(mappings[return_type].keys()))
            if perc_correct >= percentage_correct_threshold:
                new_return_types.append(return_type)

        # If no return types are left, fall-back to generic
        if len(new_return_types) == 0:

            # Map for generic
            mapping = await _define_mappings(
                data_mapping_prompt,
                mapping_type="generic",
                input_fields=list(rt.generic.keys()),
                output_fields=list(properties.keys()),
                properties=properties,
                collection_information=out,
                example_objects=subset_objects,
                settings=settings,
                lm=lm,
            )
            yield await process_update(
                message="No display types found, defined mappings for generic",
            )

            # remove any extra fields the model may have added
            mapping = {k: v for k, v in mapping.items() if k in list(rt.generic.keys())}
            mappings["generic"] = mapping

            # re-check the threshold for missing fields on generic
            num_missing = sum([m == "" for m in list(mappings[return_type].values())])
            perc_correct = 1 - (num_missing / len(mappings[return_type].keys()))
            if perc_correct >= percentage_correct_threshold:
                new_return_types = ["generic"]

        # Add the mappings to the output
        out["mappings"] = {
            return_type: mappings[return_type] for return_type in new_return_types
        }

        # always include the table return type
        out["mappings"]["table"] = {field: field for field in properties.keys()}

        # Delete existing metadata if it exists
        if await preprocessed_collection_exists_async(collection_name, client_manager):
            await delete_preprocessed_collection_async(collection_name, client_manager)

        # Save final metadata to a collection
        async with client_manager.connect_to_async_client() as client:
            if await client.collections.exists(f"ELYSIA_METADATA__"):
                metadata_collection = client.collections.get("ELYSIA_METADATA__")
            else:
                metadata_collection = await client.collections.create(
                    f"ELYSIA_METADATA__",
                    vectorizer_config=Configure.Vectorizer.none(),
                    properties=[
                        Property(
                            name="name",
                            data_type=DataType.TEXT,
                            tokenization=Tokenization.FIELD,
                        ),
                        Property(
                            name="length",
                            data_type=DataType.NUMBER,
                        ),
                        Property(
                            name="summary",
                            data_type=DataType.TEXT,
                        ),
                        Property(
                            name="index_properties",
                            data_type=DataType.OBJECT,
                            nested_properties=[
                                Property(
                                    name="isNullIndexed",
                                    data_type=DataType.BOOL,
                                ),
                                Property(
                                    name="isLengthIndexed",
                                    data_type=DataType.BOOL,
                                ),
                                Property(
                                    name="isTimestampIndexed",
                                    data_type=DataType.BOOL,
                                ),
                            ],
                        ),
                        Property(
                            name="named_vectors",
                            data_type=DataType.OBJECT_ARRAY,
                            nested_properties=[
                                Property(
                                    name="name",
                                    data_type=DataType.TEXT,
                                ),
                                Property(
                                    name="vectorizer",
                                    data_type=DataType.TEXT,
                                ),
                                Property(
                                    name="model",
                                    data_type=DataType.TEXT,
                                ),
                                Property(
                                    name="source_properties",
                                    data_type=DataType.TEXT_ARRAY,
                                ),
                                Property(
                                    name="enabled",
                                    data_type=DataType.BOOL,
                                ),
                                Property(
                                    name="description",
                                    data_type=DataType.TEXT,
                                ),
                            ],
                        ),
                        Property(
                            name="vectorizer",
                            data_type=DataType.OBJECT,
                            nested_properties=[
                                Property(name="vectorizer", data_type=DataType.TEXT),
                                Property(name="model", data_type=DataType.TEXT),
                            ],
                        ),
                        Property(
                            name="fields",
                            data_type=DataType.OBJECT_ARRAY,
                            nested_properties=[
                                Property(
                                    name="name",
                                    data_type=DataType.TEXT,
                                ),
                                Property(
                                    name="type",
                                    data_type=DataType.TEXT,
                                ),
                                Property(
                                    name="description",
                                    data_type=DataType.TEXT,
                                ),
                                Property(
                                    name="range",
                                    data_type=DataType.NUMBER_ARRAY,
                                ),
                                Property(
                                    name="date_range",
                                    data_type=DataType.DATE_ARRAY,
                                ),
                                Property(
                                    name="groups",
                                    data_type=DataType.OBJECT_ARRAY,
                                    nested_properties=[
                                        Property(
                                            name="value",
                                            data_type=DataType.TEXT,
                                        ),
                                        Property(name="count", data_type=DataType.INT),
                                    ],
                                ),
                                Property(
                                    name="date_median",
                                    data_type=DataType.DATE,
                                ),
                                Property(
                                    name="mean",
                                    data_type=DataType.NUMBER,
                                ),
                            ],
                        ),
                        # leave mappings for auto-schema generation
                    ],
                    inverted_index_config=Configure.inverted_index(
                        index_null_state=True,
                    ),
                )
            await metadata_collection.data.insert(out)

        yield await process_update(
            completed=True,
            message="Saved metadata to Weaviate",
        )

    except Exception as e:
        yield await process_update(
            error=f"Error preprocessing collection: {str(e)}",
        )

    finally:
        if close_clients_after_completion:
            await client_manager.close_clients()


async def _preprocess_async(
    collection_names: list[str] | str,
    client_manager: ClientManager | None = None,
    min_sample_size: int = 10,
    max_sample_size: int = 20,
    num_sample_tokens: int = 30000,
    settings: Settings = environment_settings,
    force: bool = False,
) -> None:

    if client_manager is None:
        client_manager = ClientManager(
            wcd_url=settings.WCD_URL, wcd_api_key=settings.WCD_API_KEY
        )
        close_clients_after_completion = True
    else:
        close_clients_after_completion = False

    if isinstance(collection_names, str):
        collection_names = [collection_names]

    try:

        if settings.LOGGING_LEVEL_INT > 20:
            for collection_name in collection_names:
                async for result in preprocess_async(
                    collection_name=collection_name,
                    client_manager=client_manager,
                    min_sample_size=min_sample_size,
                    max_sample_size=max_sample_size,
                    num_sample_tokens=num_sample_tokens,
                    force=force,
                    settings=settings,
                ):
                    if (
                        result is not None
                        and "error" in result
                        and result["error"] != ""
                    ):
                        raise Exception(result["error"])

        else:
            for collection_name in collection_names:
                with Progress() as progress:
                    task = progress.add_task(
                        f"Preprocessing {collection_name}", total=1
                    )
                    async for result in preprocess_async(
                        collection_name=collection_name,
                        client_manager=client_manager,
                        min_sample_size=min_sample_size,
                        max_sample_size=max_sample_size,
                        num_sample_tokens=num_sample_tokens,
                        force=force,
                        settings=settings,
                    ):
                        if (
                            result is not None
                            and "error" in result
                            and result["error"] != ""
                        ):
                            raise Exception(result["error"])
                        elif result is not None and "progress" in result:
                            progress.update(
                                task,
                                completed=result["progress"],
                                description=f"({collection_name}) {result['message']}",
                            )
                    progress.update(
                        task,
                        completed=1,
                        description=f"({collection_name}) Preprocessing complete!",
                    )

    finally:
        if close_clients_after_completion:
            await client_manager.close_clients()


def preprocess(
    collection_names: str | list[str],
    client_manager: ClientManager | None = None,
    min_sample_size: int = 5,
    max_sample_size: int = 100,
    num_sample_tokens: int = 30000,
    settings: Settings = environment_settings,
    force: bool = False,
) -> None:
    """
    Preprocess a collection, obtain a LLM-generated summary of the collection,
    a set of statistics for each field (such as unique categories), and a set of mappings
    from the fields to the frontend-specific fields in Elysia.

    In order:
    1. Evaluate all the data fields and groups/statistics of the data fields as a whole
    2. Write a summary of the collection via an LLM
    3. Evaluate what return types are available for this collection
    4. For each data field in the collection, evaluate what corresponding entry goes to what field in the return type (mapping)
    5. Save as a ELYSIA_METADATA__ collection

    Depending on the size of objects in the collection, you can choose the minimum and maximum sample size,
    which will be used to create a sample of objects for the LLM to create a collection summary.
    If your objects are particularly large, you can set the sample size to be smaller, to use less tokens and speed up the LLM processing.
    If your objects are small, you can set the sample size to be larger, to get a more accurate summary.
    This is a trade-off between speed/compute and accuracy.

    But note that the pre-processing step only needs to be done once for each collection.
    The output of this function is cached, so that if you run it again, it will not re-process the collection (unless the force flag is set to True).

    This function saves the output into a collection called ELYSIA_METADATA__, which is automatically called by Elysia.
    This is saved to whatever Weaviate cluster URL/API key you have configured, or in your environment variables.
    You can change this by setting the `wcd_url` and `wcd_api_key` in the settings, and pass this Settings object to this function.

    Args:
        collection_names (str | list[str]): The name(s) of the collections to preprocess.
            Can supply either a single string for one collection, or a list of strings for multiple collections.
        client_manager (ClientManager): The client manager to use.
            If not provided, a new ClientManager will be created using the environment variables/configured settings.
        min_sample_size (int): The minimum number of objects to sample from the collection to evaluate the statistics/summary. Optional, defaults to 10.
        max_sample_size (int): The maximum number of objects to sample from the collection to evaluate the statistics/summary. Optional, defaults to 20.
        num_sample_tokens (int): The maximum number of tokens in the sample objects used to evaluate the summary. Optional, defaults to 30000.
        settings (Settings): The settings to use. Optional, defaults to the environment variables/configured settings.
        force (bool): Whether to force the preprocessor to run even if the collection already exists. Optional, defaults to False.
    """

    asyncio_run(
        _preprocess_async(
            collection_names,
            client_manager,
            min_sample_size,
            max_sample_size,
            num_sample_tokens,
            settings,
            force,
        )
    )


async def preprocessed_collection_exists_async(
    collection_name: str, client_manager: ClientManager | None = None
) -> bool:
    """
    Async version of `preprocessed_collection_exists`.

    Args:
        collection_name (str): The name of the collection to check.
        client_manager (ClientManager): The client manager to use.
            If not provided, a new ClientManager will be created using the environment variables/configured settings.

    Returns:
        bool: True if the collection exists, False otherwise.
    """
    if client_manager is None:
        client_manager = ClientManager()
        close_clients_after_completion = True
    else:
        close_clients_after_completion = False

    async with client_manager.connect_to_async_client() as client:
        metadata_exists = await client.collections.exists(f"ELYSIA_METADATA__")
        if not metadata_exists:
            return False

        metadata_collection = client.collections.get("ELYSIA_METADATA__")
        metadata = await metadata_collection.query.fetch_objects(
            filters=Filter.by_property("name").equal(collection_name)
        )

    if close_clients_after_completion:
        await client_manager.close_clients()

    return metadata is not None and len(metadata.objects) > 0


def preprocessed_collection_exists(
    collection_name: str, client_manager: ClientManager | None = None
) -> bool:
    """
    Check if the preprocessed collection exists in the Weaviate cluster.
    This function simply checks if the cached preprocessed metadata exists in the Weaviate cluster.
    It does so by checking if the collection name exists in the ELYSIA_METADATA__ collection.

    Args:
        collection_name (str): The name of the collection to check.
        client_manager (ClientManager): The client manager to use.
    """
    return asyncio_run(
        preprocessed_collection_exists_async(collection_name, client_manager)
    )


async def delete_preprocessed_collection_async(
    collection_name: str, client_manager: ClientManager | None = None
) -> None:
    """
    Delete the preprocessed collection from the Weaviate cluster.
    This function simply deletes the cached preprocessed metadata from the Weaviate cluster.
    It does so by deleting the object in the collection ELYSIA_METADATA__ with the name of the collection.

    Args:
        collection_name (str): The name of the collection to delete the preprocessed metadata for.
        client_manager (ClientManager): The client manager to use.
            If not provided, a new ClientManager will be created using the environment variables/configured settings.
    """
    if client_manager is None:
        client_manager = ClientManager()
        close_clients_after_completion = True
    else:
        close_clients_after_completion = False

    async with client_manager.connect_to_async_client() as client:
        if await client.collections.exists(f"ELYSIA_METADATA__"):
            metadata_collection = client.collections.get("ELYSIA_METADATA__")
            metadata = await metadata_collection.query.fetch_objects(
                filters=Filter.by_property("name").equal(collection_name),
                limit=1,
            )
            if metadata is not None and len(metadata.objects) > 0:
                await metadata_collection.data.delete_by_id(metadata.objects[0].uuid)
            else:
                raise Exception(f"Metadata for {collection_name} does not exist")

    if close_clients_after_completion:
        await client_manager.close_clients()


def delete_preprocessed_collection(
    collection_name: str, client_manager: ClientManager | None = None
) -> None:
    """
    Delete a preprocessed collection.
    This function allows you to delete the preprocessing done for a particular collection.
    It does so by deleting the object in the ELYSIA_METADATA__ collection with the name of the collection.

    Args:
        collection_name (str): The name of the collection to delete.
        client_manager (ClientManager): The client manager to use.
    """
    return asyncio_run(
        delete_preprocessed_collection_async(collection_name, client_manager)
    )


def edit_preprocessed_collection(
    collection_name: str,
    client_manager: ClientManager | None = None,
    named_vectors: list[dict] | None = None,
    summary: str | None = None,
    mappings: dict[str, dict[str, str]] | None = None,
    fields: list[dict[str, str] | None] | None = None,
) -> None:
    """
    Edit a preprocessed collection.
    This function allows you to edit the named vectors, summary, mappings, and fields of a preprocessed collection.
    It does so by updating the ELYSIA_METADATA__ collection.
    Find available mappings in the `elysia.util.return_types` module.

    Args:
        collection_name (str): The name of the collection to edit.
        client_manager (ClientManager): The client manager to use.
            If not provided, a new ClientManager will be created using the environment variables/configured settings.
        named_vectors (list[dict]): The named vectors to update. This has fields "name", "enabled", and "description".
            The "name" is used to identify the named vector to change (the name will not change).
            Set "enabled" to True/False to enable/disable the named vector.
            Set "description" to describe the named vector.
            The description of named vectors is not automatically generated by the LLM.
            Any named vectors that are not provided will not be updated.
            If None or not provided, the named vectors will not be updated.
        summary (str): The summary to update.
            The summary is a short description of the collection, generated by the LLM.
            This will replace the existing summary of the collection.
            If None or not provided, the summary will not be updated.
        mappings (dict): The mappings to update.
            The mappings are what the frontend will use to display the collection, and the associated fields.
            I.e., which fields correspond to which output fields on the frontend.
            The keys of the outer level of the dictionary are the mapping names, the values are dictionaries with the mappings.
            The inner dictionary has the keys as the collection fields, and the values as the frontend fields.
            If None or not provided, the mappings will not be updated.
        fields (list[dict]): The fields to update.
            Each element in the list is a dictionary with the following fields:
            - "name": The name of the field. (This is used to identify the field to change, the name will not change).
            - "description": The description of the field to update.
            Any fields that are not provided will not be updated.
            If None or not provided, the fields will not be updated.

    Returns:
        dict: The updated preprocessed collection.
    """

    return asyncio_run(
        edit_preprocessed_collection_async(
            collection_name, client_manager, named_vectors, summary, mappings, fields
        )
    )


async def edit_preprocessed_collection_async(
    collection_name: str,
    client_manager: ClientManager | None = None,
    named_vectors: list[dict] | None = None,
    summary: str | None = None,
    mappings: dict[str, dict[str, str]] | None = None,
    fields: list[dict[str, str] | None] | None = None,
) -> dict:
    """
    Async version of `edit_preprocessed_collection`.

    Args:
        collection_name (str): The name of the collection to edit.
        client_manager (ClientManager): The client manager to use.
            If not provided, a new ClientManager will be created using the environment variables/configured settings.
        named_vectors (list[dict]): The named vectors to update. This has fields "name", "enabled", and "description".
            The "name" is used to identify the named vector to change (the name will not change).
            Set "enabled" to True/False to enable/disable the named vector.
            Set "description" to describe the named vector.
            The description of named vectors is not automatically generated by the LLM.
            Any named vectors that are not provided will not be updated.
            If None or not provided, the named vectors will not be updated.
        summary (str): The summary to update.
            The summary is a short description of the collection, generated by the LLM.
            This will replace the existing summary of the collection.
            If None or not provided, the summary will not be updated.
        mappings (dict): The mappings to update.
            The mappings are what the frontend will use to display the collection, and the associated fields.
            I.e., which fields correspond to which output fields on the frontend.
            The keys of the outer level of the dictionary are the mapping names, the values are dictionaries with the mappings.
            The inner dictionary has the keys as the collection fields, and the values as the frontend fields.
            If None or not provided, the mappings will not be updated.
        fields (list[dict]): The fields to update.
            Each element in the list is a dictionary with the following fields:
            - "name": The name of the field. (This is used to identify the field to change, the name will not change).
            - "description": The description of the field to update.
            Any fields that are not provided will not be updated.
            If None or not provided, the fields will not be updated.

    Returns:
        dict: The updated preprocessed collection.
    """

    if client_manager is None:
        client_manager = ClientManager()
        close_clients_after_completion = True
    else:
        close_clients_after_completion = False

    async with client_manager.connect_to_async_client() as client:
        metadata_name = f"ELYSIA_METADATA__"

        # check if the collection itself exists
        if not await client.collections.exists(collection_name):
            raise Exception(f"Collection {collection_name} does not exist")

        # check if the metadata collection exists
        if not await client.collections.exists(metadata_name):
            raise Exception(f"Metadata collection does not exist")

        else:
            metadata_collection = client.collections.get(metadata_name)
            metadata = await metadata_collection.query.fetch_objects(
                filters=Filter.by_property("name").equal(collection_name),
                limit=1,
            )
            uuid = metadata.objects[0].uuid
            properties: dict = metadata.objects[0].properties  # type: ignore

        # update the named vectors
        if named_vectors is not None:
            for named_vector in named_vectors:
                for property_named_vector in properties["named_vectors"]:
                    if property_named_vector["name"] == named_vector["name"]:

                        if named_vector["enabled"] is not None:
                            property_named_vector["enabled"] = named_vector["enabled"]

                        if named_vector["description"] is not None:
                            property_named_vector["description"] = named_vector[
                                "description"
                            ]

        # update the summary
        if summary is not None:
            properties["summary"] = summary

        # update the mappings
        if mappings is not None:
            if "table" in mappings:
                mappings["table"] = {
                    field["name"]: field["name"] for field in properties["fields"]
                }

            # format all mappings
            for mapping_type, mapping in mappings.items():
                if mapping_type == "table":
                    continue

                # check if the mapping_type is valid
                if mapping_type not in rt.types_dict:
                    raise ValueError(
                        f"Invalid mapping type: {mapping_type}. Valid mapping types are: {list(rt.types_dict.keys())}"
                    )

                # check if the mapping is valid
                for field_name, field_value in mapping.items():
                    if field_name not in rt.types_dict[mapping_type]:
                        raise ValueError(
                            f"Invalid field name: {field_name} for mapping type: {mapping_type}. "
                            f"Valid fields are: {list(rt.types_dict[mapping_type].keys())}"
                        )

                # add empty fields
                for true_field in rt.types_dict[mapping_type]:
                    if true_field not in mapping:
                        mapping[true_field] = ""

            properties["mappings"] = mappings

            # check if the `conversation_id` field is in the mapping (required for conversation type)
            if "conversation" in mappings and (
                mappings["conversation"]["conversation_id"] is None
                or mappings["conversation"]["conversation_id"] == ""
            ):
                raise ValueError(
                    "Conversation type requires a conversation_id field, but none was found in the mappings for conversation. "
                )

            if "conversation" in mappings and (
                mappings["conversation"]["message_id"] is None
                or mappings["conversation"]["message_id"] == ""
            ):
                raise ValueError(
                    "Conversation type requires a message_id field, but none was found in the mappings for conversation. "
                )

            # check if there is a message type as well as conversation
            if "message" not in mappings and "conversation" in mappings:
                raise ValueError(
                    "Conversation type requires message type to also be set as a fallback."
                )

        # update the fields
        if fields is not None:
            for field in fields:
                for property_field in properties["fields"]:
                    if field is not None and property_field["name"] == field["name"]:
                        property_field["description"] = field["description"]

        format_dict_to_serialisable(properties)

        # update the collection
        await metadata_collection.data.update(uuid=uuid, properties=properties)

    if close_clients_after_completion:
        await client_manager.close_clients()

    return properties


def view_preprocessed_collection(
    collection_name: str, client_manager: ClientManager | None = None
) -> dict:
    """
    View a preprocessed collection.
    This function allows you to view the preprocessed collection generated by the preprocess function.
    It does so by querying the ELYSIA_METADATA__ collection.

    Args:
        collection_name (str): The name of the collection to view.
        client_manager (ClientManager): The client manager to use.
            If not provided, a new ClientManager will be created using the environment variables/configured settings.

    Returns:
        dict: The preprocessed collection.
    """
    return asyncio_run(
        view_preprocessed_collection_async(collection_name, client_manager)
    )


async def view_preprocessed_collection_async(
    collection_name: str, client_manager: ClientManager | None = None
) -> dict:
    """
    Async version of `view_preprocessed_collection`.

    Args:
        collection_name (str): The name of the collection to view.
        client_manager (ClientManager): The client manager to use.
            If not provided, a new ClientManager will be created using the environment variables/configured settings.

    Returns:
        dict: The preprocessed collection.
    """

    if client_manager is None:
        client_manager = ClientManager()
        close_clients_after_completion = True
    else:
        close_clients_after_completion = False

    async with client_manager.connect_to_async_client() as client:
        metadata_name = f"ELYSIA_METADATA__"

        # check if the collection itself exists
        if not await client.collections.exists(collection_name):
            raise Exception(f"Collection {collection_name} does not exist")

        # check if the metadata collection exists
        if not await client.collections.exists(metadata_name):
            raise Exception(f"Metadata collection does not exist")

        else:
            metadata_collection = client.collections.get(metadata_name)
            metadata = await metadata_collection.query.fetch_objects(
                filters=Filter.by_property("name").equal(collection_name),
                limit=1,
            )
            if len(metadata.objects) == 0:
                raise Exception(f"Metadata for {collection_name} does not exist")

            properties: dict = metadata.objects[0].properties  # type: ignore

    if close_clients_after_completion:
        await client_manager.close_clients()

    return properties



import os
import logging
import litellm
from litellm import (
    AuthenticationError,
    NotFoundError,
    BadRequestError,
    models_by_provider,
)
from litellm.utils import get_valid_models, check_valid_key
from rich.logging import RichHandler
from typing import Callable, Literal

import spacy
import random

from dotenv import load_dotenv
from dspy import LM
from copy import deepcopy

load_dotenv(override=True)

try:
    nlp = spacy.load("en_core_web_sm")
except Exception:
    spacy.cli.download("en_core_web_sm")  # type: ignore
    nlp = spacy.load("en_core_web_sm")


api_key_to_provider = {
    "openai_api_key": ["openai"],
    "anthropic_api_key": ["anthropic"],
    "openrouter_api_key": [
        "openrouter/openai",
        "openrouter/anthropic",
        "openrouter/google",
    ],
    "gemini_api_key": ["gemini"],
    "azure_api_key": ["azure"],
}

provider_to_models = {
    "openai": ["gpt-4.1", "gpt-4.1-mini", "gpt-4.1-nano", "gpt-4o-mini", "gpt-4o"],
    "anthropic": [
        "claude-sonnet-4-20250514",
        "claude-3-7-sonnet-20250219",
        "claude-3-5-haiku-20241022",
    ],
    "openrouter/openai": [
        "gpt-4.1",
        "gpt-4.1-mini",
        "gpt-4.1-nano",
        "gpt-4o-mini",
        "gpt-4o",
    ],
    "openrouter/anthropic": [
        "claude-sonnet-4",
        "claude-3.7-sonnet",
        "claude-3.5-haiku",
    ],
    "openrouter/google": [
        "gemini-2.5-flash",
        "gemini-2.0-flash-001",
        "gemini-2.5-pro",
        "gemini-2.5-flash-lite",
    ],
    "gemini": [
        "gemini-2.5-flash",
        "gemini-2.0-flash-001",
        "gemini-2.5-pro",
        "gemini-2.5-flash-lite",
    ],
}

provider_to_api_keys = {
    "openai": ["openai_api_key"],
    "anthropic": ["anthropic_api_key"],
    "openrouter/openai": ["openrouter_api_key"],
    "openrouter/anthropic": ["openrouter_api_key"],
    "openrouter/google": ["openrouter_api_key"],
    "gemini": ["gemini_api_key"],
    "groq": ["groq_api_key"],
    "databricks": ["databricks_api_key", "databricks_api_base"],
    "bedrock": ["aws_access_key_id", "aws_secret_access_key", "aws_region_name"],
    "snowflake": ["snowflake_jwt", "snowflake_account_id"],
}


def get_available_models(api_keys: list[str]):
    available_models = []
    for api_key in api_keys:
        if api_key in api_key_to_provider:
            for provider in api_key_to_provider[api_key]:
                if provider in provider_to_models:
                    available_models.extend(provider_to_models[provider])
    return list(set(available_models))


def get_available_providers(api_keys: list[str]) -> list[str]:
    available_providers = []
    for api_key in api_keys:
        if api_key in api_key_to_provider:
            available_providers.extend(api_key_to_provider[api_key])
    return list(set(available_providers))


def is_api_key(key: str) -> bool:
    return (
        key.lower().endswith("api_key")
        or key.lower().endswith("apikey")
        or key.lower().endswith("api_version")
        or key.lower().endswith("api_base")
        or key.lower().endswith("_account_id")
        or key.lower().endswith("_jwt")
        or key.lower().endswith("_access_key_id")
        or key.lower().endswith("_secret_access_key")
        or key.lower().endswith("_region_name")
    )


class Settings:
    """
    Settings for Elysia.
    This class handles the configuration of various settings within Elysia.
    This includes:
    - The base and complex models to use.
    - The providers for the base and complex models.
    - The Weaviate cloud URL and API key.
    - The API keys for the providers.
    - The logger and logging level.

    The Settings object is set as at a default `settings` object if running Elysia as a package.
    You can import this via `elysia.config.settings`.
    This is initialised to default values from the environment variables.

    Or, you can create your own Settings object, and configure it as you wish.
    The Settings object can be passed to different Elysia classes and functions, such as `Tree` and `preprocess`.
    These will not use the global `settings` object, but instead use the Settings object you passed to them.
    """

    def __init__(self):
        """
        Initialize the settings for Elysia.
        These are all settings initialised to None, and should be set using the `configure` method.
        """
        # Default settings
        self.SETTINGS_ID = str(random.randint(100000000000000, 999999999999999))

        self.base_init()

    def base_init(self):
        self.BASE_MODEL: str | None = None
        self.BASE_PROVIDER: str | None = None
        self.COMPLEX_MODEL: str | None = None
        self.COMPLEX_PROVIDER: str | None = None

        self.WCD_URL: str = ""
        self.WCD_API_KEY: str = ""

        self.MODEL_API_BASE: str | None = None

        self.API_KEYS: dict[str, str] = {}

        self.logger = logging.getLogger("rich")
        self.logger.setLevel(logging.INFO)

        # Remove any existing handlers before adding a new one
        for handler in self.logger.handlers[:]:
            self.logger.removeHandler(handler)
        self.logger.addHandler(RichHandler(rich_tracebacks=True, markup=True))

        self.logger.propagate = False
        self.LOGGING_LEVEL = "INFO"
        self.LOGGING_LEVEL_INT = 20

        # Experimental features
        self.USE_FEEDBACK = False
        self.BASE_USE_REASONING = True
        self.COMPLEX_USE_REASONING = True

    def setup_app_logger(self, logger: logging.Logger):
        """
        Override existing logger with the app-level logger.

        Args:
            logger (Logger): The logger to use.
        """
        self.logger = logger
        self.LOGGING_LEVEL_INT = logger.level
        inverted_logging_mapping = {
            v: k for k, v in logging.getLevelNamesMapping().items()
        }
        self.LOGGING_LEVEL = inverted_logging_mapping[self.LOGGING_LEVEL_INT]

    def configure_logger(
        self,
        level: Literal[
            "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "NOTSET"
        ] = "NOTSET",
    ):
        """
        Configure the logger with a RichHandler.

        Args:
            level (Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "NOTSET"]): The logging level to use.
        """
        self.logger.setLevel(level)
        self.LOGGING_LEVEL = level
        self.LOGGING_LEVEL_INT = logging.getLevelNamesMapping()[level]

    def set_api_key(self, api_key: str, api_key_name: str) -> None:
        self.API_KEYS[api_key_name] = api_key

    def get_api_key(self, api_key_name: str) -> str:
        return self.API_KEYS[api_key_name]

    def load_settings(self, settings: dict):
        for item in settings:
            setattr(self, item, settings[item])

        # self.logger = logging.getLogger("rich")
        # self.logger.setLevel(self.LOGGING_LEVEL)
        # for handler in self.logger.handlers[:]:
        #     self.logger.removeHandler(handler)
        # self.logger.addHandler(RichHandler(rich_tracebacks=True, markup=True))

    @classmethod
    def from_smart_setup(cls):
        settings = cls()
        settings.set_from_env()
        settings.smart_setup()
        return settings

    @classmethod
    def from_env_vars(cls):
        settings = cls()
        settings.set_from_env()
        return settings

    def set_from_env(self):
        self.BASE_MODEL = os.getenv("BASE_MODEL", None)
        self.COMPLEX_MODEL = os.getenv("COMPLEX_MODEL", None)
        self.BASE_PROVIDER = os.getenv("BASE_PROVIDER", None)
        self.COMPLEX_PROVIDER = os.getenv("COMPLEX_PROVIDER", None)
        self.MODEL_API_BASE = os.getenv("MODEL_API_BASE", None)
        self.LOGGING_LEVEL = os.getenv("LOGGING_LEVEL", "NOTSET")
        self.set_api_keys_from_env()

    def set_api_keys_from_env(self):

        self.WCD_URL = os.getenv(
            "WEAVIATE_URL",
            os.getenv("WCD_URL", ""),
        )
        self.WCD_API_KEY = os.getenv(
            "WEAVIATE_API_KEY",
            os.getenv("WCD_API_KEY", ""),
        )

        self.API_KEYS = {
            env_var.lower(): os.getenv(env_var, "")
            for env_var in os.environ
            if is_api_key(env_var) and env_var.lower() != "wcd_api_key"
        }
        for api_key in self.API_KEYS:
            self.set_api_key(self.API_KEYS[api_key], api_key)

    def smart_setup(self):

        # Check if the user has set the base model etc from the environment variables
        if os.getenv("BASE_MODEL", None):
            self.BASE_MODEL = os.getenv("BASE_MODEL")
        if os.getenv("COMPLEX_MODEL", None):
            self.COMPLEX_MODEL = os.getenv("COMPLEX_MODEL")
        if os.getenv("BASE_PROVIDER", None):
            self.BASE_PROVIDER = os.getenv("BASE_PROVIDER")
        if os.getenv("COMPLEX_PROVIDER", None):
            self.COMPLEX_PROVIDER = os.getenv("COMPLEX_PROVIDER")

        self.MODEL_API_BASE = os.getenv("MODEL_API_BASE", None)
        self.LOGGING_LEVEL = os.getenv("LOGGING_LEVEL", "NOTSET")

        self.set_api_keys_from_env()

        # check what API keys are available
        if (
            self.BASE_MODEL is None
            or self.COMPLEX_MODEL is None
            or self.BASE_PROVIDER is None
            or self.COMPLEX_PROVIDER is None
        ):
            if os.getenv("OPENROUTER_API_KEY", None):
                # use gemini 2.0 flash
                self.BASE_PROVIDER = "openrouter/google"
                self.COMPLEX_PROVIDER = "openrouter/google"
                self.BASE_MODEL = "gemini-2.0-flash-001"
                self.COMPLEX_MODEL = "gemini-2.5-flash"
            elif os.getenv("GEMINI_API_KEY", None):
                # use gemini 2.0 flash
                self.BASE_PROVIDER = "gemini"
                self.COMPLEX_PROVIDER = "gemini"
                self.BASE_MODEL = "gemini-2.0-flash-001"
                self.COMPLEX_MODEL = "gemini-2.5-flash"
            elif os.getenv("OPENAI_API_KEY", None):
                # use gpt family
                self.BASE_PROVIDER = "openai"
                self.COMPLEX_PROVIDER = "openai"
                self.BASE_MODEL = "gpt-4.1-mini"
                self.COMPLEX_MODEL = "gpt-4.1"
            elif os.getenv("ANTHROPIC_API_KEY", None):
                # use claude family
                self.BASE_PROVIDER = "anthropic"
                self.COMPLEX_PROVIDER = "anthropic"
                self.BASE_MODEL = "claude-3-5-haiku-latest"
                self.COMPLEX_MODEL = "claude-sonnet-4-0"

    def configure(
        self,
        replace: bool = False,
        **kwargs,
    ):
        """
        Configure the settings for Elysia for the current Settings object.

        Args:
            replace (bool): Whether to override the current settings with the new settings.
                When this is True, all existing settings are removed, and only the new settings are used.
                Defaults to False.
            **kwargs (str): One or more of the following:
                - base_model (str): The base model to use. e.g. "gpt-4o-mini"
                - complex_model (str): The complex model to use. e.g. "gpt-4o"
                - base_provider (str): The provider to use for base_model. E.g. "openai" or "openrouter/openai"
                - complex_provider (str): The provider to use for complex_model. E.g. "openai" or "openrouter/openai"
                - model_api_base (str): The API base to use.
                - wcd_url (str): The Weaviate cloud URL to use.
                - wcd_api_key (str): The Weaviate cloud API key to use.
                - logging_level (str): The logging level to use. e.g. "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"
                - use_feedback (bool): EXPERIMENTAL. Whether to use feedback from previous runs of the tree.
                    If True, the tree will use TrainingUpdate objects that have been saved in previous runs of the decision tree.
                    These are implemented via few-shot examples for the decision node.
                    They are collected in the 'feedback' collection (ELYSIA_FEEDBACK__).
                    Relevant examples are retrieved from the collection based on searching the collection via the user's prompt.
                - base_use_reasoning (bool): Whether to use reasoning output for the base model.
                    If True, the model will generate reasoning before coming to its solution.
                - complex_use_reasoning (bool): Whether to use reasoning output for the complex model.
                    If True, the model will generate reasoning before coming to its solution.
                - Additional API keys to set. E.g. `openai_apikey="..."`, if this argument ends with `apikey` or `api_key`,
                    it will be added to the `API_KEYS` dictionary.

        """
        if replace:
            self.base_init()

        # convert all kwargs to lowercase for consistency
        kwargs = {kwarg.lower(): kwargs[kwarg] for kwarg in kwargs}

        if "base_model" in kwargs:
            if "base_provider" not in kwargs:
                raise ValueError(
                    "Provider must be specified if base_model is set. "
                    "E.g. `elysia.config.configure(base_model='gpt-4o-mini', base_provider='openai')`"
                )

            if kwargs["base_provider"] == "ollama" and (
                not kwargs["model_api_base"] or "MODEL_API_BASE" not in dir(self)
            ):
                raise ValueError(
                    "Using local models via ollama requires MODEL_API_BASE to be set. "
                    "This is likely to be http://localhost:11434. "
                    "e.g. `elysia.config.configure(model_api_base='http://localhost:11434')`"
                )

            self.BASE_MODEL = kwargs["base_model"]
            self.BASE_PROVIDER = kwargs["base_provider"]

            kwargs.pop("base_model")
            kwargs.pop("base_provider")

            # self.load_base_dspy_model()

        if "complex_model" in kwargs:
            if "complex_provider" not in kwargs:
                raise ValueError(
                    "Provider must be specified if complex_model is set. "
                    "E.g. `elysia.config.configure(complex_model='gpt-4o', complex_provider='openai')`"
                )

            if kwargs["complex_provider"] == "ollama" and (
                not kwargs["model_api_base"] or "MODEL_API_BASE" not in dir(self)
            ):
                raise ValueError(
                    "Using local models via ollama requires MODEL_API_BASE to be set. "
                    "This is likely to be http://localhost:11434. "
                    "e.g. `elysia.config.configure(model_api_base='http://localhost:11434')`"
                )

            self.COMPLEX_MODEL = kwargs["complex_model"]
            self.COMPLEX_PROVIDER = kwargs["complex_provider"]

            kwargs.pop("complex_model")
            kwargs.pop("complex_provider")

            # self.load_complex_dspy_model()

        if "model_api_base" in kwargs:
            self.MODEL_API_BASE = kwargs["model_api_base"]
            kwargs.pop("model_api_base")

        if "wcd_url" in kwargs:
            self.WCD_URL = kwargs["wcd_url"]
            kwargs.pop("wcd_url")

        if "wcd_api_key" in kwargs:
            self.WCD_API_KEY = kwargs["wcd_api_key"]
            kwargs.pop("wcd_api_key")

        if "weaviate_url" in kwargs:
            self.WCD_URL = kwargs["weaviate_url"]
            kwargs.pop("weaviate_url")

        if "weaviate_api_key" in kwargs:
            self.WCD_API_KEY = kwargs["weaviate_api_key"]
            kwargs.pop("weaviate_api_key")

        if "logging_level" in kwargs or "logger_level" in kwargs:

            self.LOGGING_LEVEL = (
                kwargs["logging_level"]
                if "logging_level" in kwargs
                else kwargs["logger_level"]
            )
            self.LOGGING_LEVEL_INT = logging.getLevelNamesMapping()[self.LOGGING_LEVEL]
            self.logger.setLevel(self.LOGGING_LEVEL)
            if "logging_level" in kwargs:
                kwargs.pop("logging_level")
            if "logger_level" in kwargs:
                kwargs.pop("logger_level")
            if "logging_level_int" in kwargs:
                kwargs.pop("logging_level_int")
            if "logger_level_int" in kwargs:
                kwargs.pop("logger_level_int")

        if "logging_level_int" in kwargs or "logger_level_int" in kwargs:

            self.LOGGING_LEVEL_INT = (
                kwargs["logging_level_int"]
                if "logging_level_int" in kwargs
                else kwargs["logger_level_int"]
            )
            self.LOGGING_LEVEL = {
                v: k for k, v in logging.getLevelNamesMapping().items()
            }[self.LOGGING_LEVEL_INT]
            self.logger.setLevel(self.LOGGING_LEVEL)

        if "settings_id" in kwargs:
            self.SETTINGS_ID = kwargs["settings_id"]
            kwargs.pop("settings_id")

        if "use_feedback" in kwargs:
            self.USE_FEEDBACK = kwargs["use_feedback"]
            kwargs.pop("use_feedback")

        if "base_use_reasoning" in kwargs:
            self.BASE_USE_REASONING = kwargs["base_use_reasoning"]
            kwargs.pop("base_use_reasoning")

        if "complex_use_reasoning" in kwargs:
            self.COMPLEX_USE_REASONING = kwargs["complex_use_reasoning"]
            kwargs.pop("complex_use_reasoning")

        if "api_keys" in kwargs and isinstance(kwargs["api_keys"], dict):
            for key, value in kwargs["api_keys"].items():
                self.set_api_key(value, key)
            kwargs.pop("api_keys")

        # remainder of kwargs are API keys or saved there
        removed_kwargs = []
        for key, value in kwargs.items():
            if is_api_key(key):
                self.set_api_key(value, key)
                removed_kwargs.append(key)

        for key in removed_kwargs:
            kwargs.pop(key)

        # remaining kwargs are not API keys
        if len(kwargs) > 0:
            self.logger.warning(
                "Unknown arguments to configure: " + ", ".join(kwargs.keys())
            )

    def __repr__(self) -> str:
        out = ""
        if "BASE_MODEL" in dir(self) and self.BASE_MODEL is not None:
            out += f"Base model: {self.BASE_MODEL}\n"
        else:
            out += "Base model: not set\n"
        if "COMPLEX_MODEL" in dir(self) and self.COMPLEX_MODEL is not None:
            out += f"Complex model: {self.COMPLEX_MODEL}\n"
        else:
            out += "Complex model: not set\n"
        if "BASE_PROVIDER" in dir(self) and self.BASE_PROVIDER is not None:
            out += f"Base provider: {self.BASE_PROVIDER}\n"
        else:
            out += "Base provider: not set\n"
        if "COMPLEX_PROVIDER" in dir(self) and self.COMPLEX_PROVIDER is not None:
            out += f"Complex provider: {self.COMPLEX_PROVIDER}\n"
        else:
            out += "Complex provider: not set\n"
        if "MODEL_API_BASE" in dir(self):
            out += f"Model API base: {self.MODEL_API_BASE}\n"
        else:
            out += "Model API base: not set\n"
        return out

    def to_json(self):
        return {
            item: getattr(self, item)
            for item in dir(self)
            if not item.startswith("_")
            and not isinstance(getattr(self, item), Callable)
            and item not in ["BASE_MODEL_LM", "COMPLEX_MODEL_LM", "logger"]
        }

    @classmethod
    def from_json(cls, json_data: dict):
        settings = cls()
        for item in json_data:
            if item not in ["logger"]:
                setattr(settings, item, json_data[item])

        settings.logger.setLevel(settings.LOGGING_LEVEL)

        return settings

    def check(self):
        return {
            "base_model": self.BASE_MODEL is not None and self.BASE_MODEL != "",
            "base_provider": self.BASE_PROVIDER is not None
            and self.BASE_PROVIDER != "",
            "complex_model": self.COMPLEX_MODEL is not None
            and self.COMPLEX_MODEL != "",
            "complex_provider": self.COMPLEX_PROVIDER is not None
            and self.COMPLEX_PROVIDER != "",
            "wcd_url": self.WCD_URL != "",
            "wcd_api_key": self.WCD_API_KEY != "",
        }


class APIKeyError(Exception):
    pass


class IncorrectModelError(Exception):
    pass


class ElysiaKeyManager:
    def __init__(self, settings: Settings):
        self.settings = settings

    def _check_model_availability(self, model: str, provider: str) -> bool:
        if provider not in models_by_provider and not provider.startswith("openrouter"):
            raise IncorrectModelError(
                f"The provider {provider} is not available. "
                f"Some example providers are: {', '.join(list(models_by_provider.keys())[:5])}. "
                "Check the full model documentation (https://docs.litellm.ai/docs/providers)."
            )

        elif (
            provider.startswith("openrouter/")
            and f"{provider}/{model}" not in models_by_provider["openrouter"]
        ):
            example_models = [
                model.split("/")[-1]
                for model in models_by_provider["openrouter"]
                if model.startswith(provider)
            ]

            raise IncorrectModelError(
                f"The model {model} is not available for provider {provider}. "
                f"Some example models for {provider} are: "
                f"{', '.join(example_models[:5])}. "
                "Check the full model documentation (https://docs.litellm.ai/docs/providers)."
            )

        elif (
            provider in models_by_provider and model not in models_by_provider[provider]
        ):
            raise IncorrectModelError(
                f"The model {model} is not available for provider {provider}. "
                f"Some example models for {provider} are: "
                f"{', '.join(models_by_provider[provider][:5])}. "
                "Check the full model documentation (https://docs.litellm.ai/docs/providers)."
            )

        return False

    def __enter__(self):

        # save existing env
        self.existing_env = deepcopy(os.environ)

        # blank out the environ except for certain values
        os.environ = {
            "MODEL_API_BASE": self.settings.MODEL_API_BASE,
            "WCD_URL": self.settings.WCD_URL,
            "WCD_API_KEY": self.settings.WCD_API_KEY,
        }

        # update all api keys in env
        for api_key, value in self.settings.API_KEYS.items():
            os.environ[api_key.upper()] = value

        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        # reset env to original
        os.environ = self.existing_env

        if exc_type is NotFoundError or exc_type is BadRequestError:
            self._check_model_availability(
                self.settings.BASE_MODEL, self.settings.BASE_PROVIDER
            )
            self._check_model_availability(
                self.settings.COMPLEX_MODEL, self.settings.COMPLEX_PROVIDER
            )

        if exc_type is AuthenticationError or exc_type is BadRequestError:
            missing_base_api_keys = [
                api_key
                for api_key in provider_to_api_keys[self.settings.BASE_PROVIDER]
                if api_key not in self.settings.API_KEYS
            ]
            missing_complex_api_keys = [
                api_key
                for api_key in provider_to_api_keys[self.settings.COMPLEX_PROVIDER]
                if api_key not in self.settings.API_KEYS
            ]
            if len(missing_base_api_keys) > 0:
                raise APIKeyError(
                    f"You are trying to use the model '{self.settings.BASE_MODEL}' "
                    f"but you do not have one of the following API keys: {', '.join(missing_base_api_keys)}. "
                    f"Please update your API keys in the settings."
                )
            if len(missing_complex_api_keys) > 0:
                raise APIKeyError(
                    f"You are trying to use the model '{self.settings.COMPLEX_MODEL}' "
                    f"but you do not have one of the following API keys: {', '.join(missing_complex_api_keys)}. "
                    f"Please update your API keys in the settings."
                )

            raise APIKeyError(
                f"One of your API keys is incorrect. "
                f"Please update your API keys in the settings. "
                f"The relevant API keys are: "
                f"{set(provider_to_api_keys[self.settings.BASE_PROVIDER] + provider_to_api_keys[self.settings.COMPLEX_PROVIDER])}"
            )

        if (
            exc_type is AuthenticationError
            or exc_type is BadRequestError
            or exc_type is NotFoundError
        ):
            raise IncorrectModelError(
                f"Either one of the models or providers: '{self.settings.BASE_MODEL}' or '{self.settings.COMPLEX_MODEL}' "
                f"({self.settings.BASE_PROVIDER} or {self.settings.COMPLEX_PROVIDER}) "
                f"is incorrect, or you do not have the required API keys. "
                f"Check the model documentation (https://docs.litellm.ai/docs/providers)."
            )

        pass


def check_base_lm_settings(settings: Settings):
    if "BASE_MODEL" not in dir(settings) or settings.BASE_MODEL is None:
        available_models = get_available_models(list(settings.API_KEYS.keys()))
        raise IncorrectModelError(
            f"No base model specified. "
            f"Available models based on your API keys: {', '.join(available_models)}"
        )

    if "BASE_PROVIDER" not in dir(settings) or settings.BASE_PROVIDER is None:
        available_providers = get_available_providers(list(settings.API_KEYS.keys()))
        raise IncorrectModelError(
            f"No base provider specified. "
            f"Available providers based on your API keys: {', '.join(available_providers)}"
        )


def check_complex_lm_settings(settings: Settings):
    if "COMPLEX_MODEL" not in dir(settings) or settings.COMPLEX_MODEL is None:
        available_models = get_available_models(list(settings.API_KEYS.keys()))
        raise IncorrectModelError(
            f"No complex model specified. "
            f"Available models based on your API keys: {', '.join(available_models)}"
        )

    if "COMPLEX_PROVIDER" not in dir(settings) or settings.COMPLEX_PROVIDER is None:
        available_providers = get_available_providers(list(settings.API_KEYS.keys()))
        raise IncorrectModelError(
            f"No complex provider specified. "
            f"Available providers based on your API keys: {', '.join(available_providers)}"
        )


def load_base_lm(settings: Settings) -> LM:
    check_base_lm_settings(settings)

    return load_lm(
        settings.BASE_PROVIDER,
        settings.BASE_MODEL,
        settings.MODEL_API_BASE if "MODEL_API_BASE" in dir(settings) else None,
    )


def load_complex_lm(settings: Settings) -> LM:
    check_complex_lm_settings(settings)

    return load_lm(
        settings.COMPLEX_PROVIDER,
        settings.COMPLEX_MODEL,
        settings.MODEL_API_BASE if "MODEL_API_BASE" in dir(settings) else None,
    )


def load_lm(
    provider: str | None,
    lm_name: str | None,
    model_api_base: str | None = None,
) -> LM:

    if provider is None or lm_name is None:
        raise ValueError("Provider and LM name must be set")

    api_base = model_api_base if provider == "ollama" else None
    full_lm_name = f"{provider}/{lm_name}"

    if lm_name.startswith("o1") or lm_name.startswith("o3"):
        return LM(
            model=full_lm_name, api_base=api_base, max_tokens=8000, temperature=1.0
        )

    return LM(model=full_lm_name, api_base=api_base, max_tokens=8000)


# global settings that should never be used by the frontend
# but used when using Elysia as a package
settings = Settings()
settings.smart_setup()

DEFAULT_SETTINGS = Settings()
DEFAULT_SETTINGS.smart_setup()


def smart_setup() -> None:
    settings.smart_setup()


def set_from_env() -> None:
    settings.set_from_env()


def reset_settings() -> None:
    settings.base_init()
    settings.smart_setup()


def configure(**kwargs) -> None:
    """
    Configure the settings for Elysia for the global settings object.

    Args:
        **kwargs (str): One or more of the following:
            - base_model (str): The base model to use. e.g. "gpt-4o-mini"
            - complex_model (str): The complex model to use. e.g. "gpt-4o"
            - base_provider (str): The provider to use for base_model. E.g. "openai" or "openrouter/openai"
            - complex_provider (str): The provider to use for complex_model. E.g. "openai" or "openrouter/openai"
            - model_api_base (str): The API base to use.
            - wcd_url (str): The Weaviate cloud URL to use.
            - wcd_api_key (str): The Weaviate cloud API key to use.
            - logging_level (str): The logging level to use. e.g. "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"
            - use_feedback (bool): EXPERIMENTAL. Whether to use feedback from previous runs of the tree.
                If True, the tree will use TrainingUpdate objects that have been saved in previous runs of the decision tree.
                These are implemented via few-shot examples for the decision node.
                They are collected in the 'feedback' collection (ELYSIA_FEEDBACK__).
                Relevant examples are retrieved from the collection based on searching the collection via the user's prompt.
            - Additional API keys to set. E.g. `openai_apikey="..."`, if this argument ends with `apikey` or `api_key`,
                it will be added to the `API_KEYS` dictionary.

    """
    settings.configure(**kwargs)



import os
from cryptography.fernet import InvalidToken
import json
from pathlib import Path

from fastapi import APIRouter, Depends
from fastapi.responses import JSONResponse
from dotenv import load_dotenv, set_key

load_dotenv(override=True)

from elysia.api.api_types import (
    SaveConfigUserData,
    UpdateFrontendConfigData,
)
from elysia.api.utils.config import Config
from elysia.api.core.log import logger
from elysia.api.dependencies.common import get_user_manager
from elysia.api.services.user import UserManager
from elysia.util.parsing import format_dict_to_serialisable, format_datetime
from elysia.config import Settings
from elysia.api.services.tree import TreeManager
from elysia.api.utils.config import FrontendConfig
from elysia.api.utils.encryption import encrypt_api_keys, decrypt_api_keys
from elysia.api.utils.models import models

import weaviate.classes.config as wc
from weaviate.util import generate_uuid5
from weaviate.classes.query import MetadataQuery, Sort, Filter
from weaviate.classes.config import Property, DataType
from uuid import uuid4


def rename_keys(config_item: dict):

    renamed_settings = {
        key.upper(): config_item["settings"][key] for key in config_item["settings"]
    }

    if "API_KEYS" in config_item["settings"]:
        renamed_api_keys = {
            key.lower(): config_item["settings"]["API_KEYS"][key]
            for key in config_item["settings"]["API_KEYS"]
        }
        renamed_settings["API_KEYS"] = renamed_api_keys

    renamed_config = {key.lower(): config_item[key] for key in config_item}
    renamed_config["settings"] = renamed_settings

    return renamed_config


router = APIRouter()


@router.get("/models")
async def get_models():
    try:
        return JSONResponse(content={"models": models, "error": ""})
    except Exception as e:
        logger.exception(f"Error in /get_models API")
        return JSONResponse(content={"models": {}, "error": str(e)})


@router.patch("/frontend_config/{user_id}")
async def update_frontend_config(
    user_id: str,
    data: UpdateFrontendConfigData,
    user_manager: UserManager = Depends(get_user_manager),
):
    """
    Update a user's frontend config, e.g. whether to save trees to weaviate or not, and the save location.
    """
    logger.debug(f"/update_frontend_config API request received")
    logger.debug(f"User ID: {user_id}")
    logger.debug(f"Config: {data.config}")

    try:
        await user_manager.update_frontend_config(user_id, data.config)
        return JSONResponse(content={"error": ""})
    except Exception as e:

        if "Could not connect to Weaviate" in str(e):
            return JSONResponse(
                content={
                    "error": "Could not connect to Weaviate. Double check that the WCD_URL and WCD_API_KEY are correct."
                }
            )
        else:
            logger.exception(f"Error in /update_frontend_config API")
            return JSONResponse(content={"error": str(e)})


@router.get("/{user_id}")
async def get_current_user_config(
    user_id: str,
    user_manager: UserManager = Depends(get_user_manager),
):
    """
    Get the current config for a single user (in memory).
    """
    logger.debug(f"/get_current_user_config API request received")
    logger.debug(f"User ID: {user_id}")

    headers = {"Cache-Control": "no-cache"}

    try:
        user = await user_manager.get_user_local(user_id)
        config = user["tree_manager"].config.to_json()
        frontend_config = user["frontend_config"].to_json()

    except Exception as e:
        logger.exception(f"Error in /get_current_user_config API")
        return JSONResponse(
            content={"error": str(e), "config": {}, "frontend_config": {}},
            headers=headers,
        )

    return JSONResponse(
        content={
            "error": "",
            "config": config,
            "frontend_config": frontend_config,
        },
        headers=headers,
    )


@router.get("/{user_id}/{config_id}/load")
async def load_a_config(
    user_id: str,
    config_id: str,
    user_manager: UserManager = Depends(get_user_manager),
):
    """
    Get the current config for a single user (in memory).
    """
    logger.debug(f"/get_current_user_config API request received")
    logger.debug(f"User ID: {user_id}")

    headers = {"Cache-Control": "no-cache"}
    warnings = []

    try:
        user = await user_manager.get_user_local(user_id)
        frontend_config = user["frontend_config"]

        # check if the user has a valid save location
        if (
            frontend_config.save_location_wcd_url == ""
            or frontend_config.save_location_wcd_api_key == ""
        ):
            raise Exception("WCD URL or API key not found.")

        if (
            frontend_config.save_location_wcd_url == ""
            or frontend_config.save_location_wcd_api_key == ""
        ):
            raise Exception(
                "No valid destination for config load location found. "
                "Please update the save location using the /update_save_location API."
            )

        # Retrieve the config from the weaviate database
        async with (
            frontend_config.save_location_client_manager.connect_to_async_client() as client
        ):
            uuid = generate_uuid5(config_id)
            collection = client.collections.get("ELYSIA_CONFIG__")
            config_response = await collection.query.fetch_object_by_id(uuid=uuid)
            if config_response is None:
                raise Exception(f"Config with ID {config_id} not found.")

            config_item: dict = config_response.properties  # type: ignore

        format_dict_to_serialisable(config_item)

        if "settings" in config_item:
            try:
                if "API_KEYS" in config_item["settings"]:
                    if "null" in config_item["settings"]["API_KEYS"]:
                        del config_item["settings"]["API_KEYS"]["null"]

                config_item["settings"] = decrypt_api_keys(config_item["settings"])
            except InvalidToken:
                warnings.append(
                    "Invalid token for config. You will need to re-enter your API keys in the settings."
                )
                config_item["settings"]["API_KEYS"] = {}
                config_item["settings"]["WCD_API_KEY"] = ""

        # Rename the keys to the correct format
        renamed_config = rename_keys(config_item)
        if "config_id" in renamed_config:
            renamed_config["id"] = renamed_config["config_id"]
            del renamed_config["config_id"]

    except Exception as e:
        logger.exception(f"Error in /get_current_user_config API")
        return JSONResponse(
            content={"error": str(e), "config": {}, "frontend_config": {}},
            headers=headers,
        )

    return JSONResponse(
        content={
            "error": "",
            "config": renamed_config,
            "frontend_config": frontend_config.to_json(),
            "warnings": warnings,
        },
        headers=headers,
    )


@router.post("/{user_id}/new")
async def new_user_config(
    user_id: str,
    user_manager: UserManager = Depends(get_user_manager),
):
    """
    Set the base_model and complex_model to the default values (gemini 2.0 flash or environment variables), for an entire user.
    Any changes will propagate to all trees associated with the user that were not initialised with their own settings.

    Args:
        user_id (str): The user ID.
        user_manager (UserManager): The user manager instance.

    Returns:
        JSONResponse: A JSON response with the following attributes:
            error (str): An error message. Empty if no error.
            config (dict): A dictionary of the new config values.
    """
    logger.debug(f"/new_user_config API request received")
    logger.debug(f"User ID: {user_id}")

    try:
        user = await user_manager.get_user_local(user_id)
        tree_manager: TreeManager = user["tree_manager"]
        frontend_config: FrontendConfig = user["frontend_config"]

        # get existing config

        settings = Settings()
        settings.smart_setup()

        config = Config(
            id=str(uuid4()),
            name="New Config",
            settings=settings,
            style="Informative, polite and friendly.",
            agent_description="You search and query Weaviate to satisfy the user's query, providing a concise summary of the results.",
            end_goal=(
                "You have satisfied the user's query, and provided a concise summary of the results. "
                "Or, you have exhausted all options available, or asked the user for clarification."
            ),
            branch_initialisation="one_branch",
        )

        frontend_config.save_location_wcd_url = frontend_config.save_location_wcd_url
        frontend_config.save_location_wcd_api_key = (
            frontend_config.save_location_wcd_api_key
        )
        frontend_config.config = {
            "save_trees_to_weaviate": True,
            "save_configs_to_weaviate": True,
            "client_timeout": int(os.getenv("CLIENT_TIMEOUT", 3)),
            "tree_timeout": int(os.getenv("TREE_TIMEOUT", 10)),
        }

    except Exception as e:
        logger.exception(f"Error in /new_user_config API")
        return JSONResponse(
            content={
                "error": str(e),
                "config": {},
                "frontend_config": {},
            }
        )

    return JSONResponse(
        content={
            "error": "",
            "config": config.to_json(),
            "frontend_config": user["frontend_config"].to_json(),
        }
    )


def save_frontend_config_to_file(user_id: str, frontend_config: dict):

    # save frontend config locally
    try:
        elysia_package_dir = Path(__file__).parent.parent.parent  # Gets to elysia/
        config_dir = elysia_package_dir / "api" / "user_configs"
        config_dir.mkdir(parents=True, exist_ok=True)
        config_file = config_dir / f"frontend_config_{user_id}.json"

        with open(config_file, "w") as f:
            json.dump(frontend_config, f)
        logger.debug(f"Frontend config saved to local file: {config_file}")
    except Exception as e:
        logger.error(f"Error in save_frontend_config_to_file: {str(e)}")
        pass


@router.post("/{user_id}/{config_id}")
async def save_config_user(
    user_id: str,
    config_id: str,
    data: SaveConfigUserData,
    user_manager: UserManager = Depends(get_user_manager),
):
    """
    Change the global config for a user.
    The `config` dictionary can contain any of the `settings` (dict), style, agent_description, end_goal and branch_initialisation.
    You do not need to specify all the `settings`, only the ones you wish to change.

    Any changes to the _settings_ will propagate to all trees associated with the user that were not initialised with their own settings,
    as well as update the ClientManager to use the new API keys.
    However, changes to the _style_, _agent_description_, _end_goal_ and _branch_initialisation_ will not be changed for existing trees.
    I.e., only changes to API keys and model choices will be changed for existing trees (Settings).

    Additionally saves a config to the weaviate database.
    If the collection ELYSIA_CONFIG__ does not exist, it will be created.
    Uses the WCD_URL and WCD_API_KEY from /update_save_location to connect to the weaviate database.
    (These are initialised to the environment variables).

    Args:
        user_id (str): Required. The user ID.
        data (SaveConfigUserData): Required. Containing the following attributes:
            name (str): Optional. The name of the config.
            settings (dict): Optional. A dictionary of config values to set, follows the same format as the Settings.configure() method.
            style (str): Optional. The writing style of the agent.
            agent_description (str): Optional. The description of the agent.
            end_goal (str): Optional. The end goal of the agent.
            branch_initialisation (str): Optional. The branch initialisation of the agent.
        user_manager (UserManager): The user manager instance.

    Returns:
        JSONResponse: A JSON response with the following attributes:
            error (str): An error message. Empty if no error.
            config (dict): A dictionary of the updated config values.
    """
    logger.debug(f"/save_config_user API request received")
    logger.debug(f"User ID: {user_id}")
    logger.debug(f"Config ID: {config_id}")
    logger.debug(f"Name: {data.name}")
    logger.debug(f"Backend Config: {data.config}")
    logger.debug(f"Frontend config: {data.frontend_config}")
    logger.debug(f"Default: {data.default}")

    warnings = []

    try:
        user = await user_manager.get_user_local(user_id)
        tree_manager: TreeManager = user["tree_manager"]
    except Exception as e:
        logger.exception(f"Error in /save_config_user API during user retrieval")
        return JSONResponse(
            content={"error": str(e), "config": {}, "frontend_config": {}}
        )

    if data.default:

        # Save the config in memory for this user (Apply step)
        try:
            await user_manager.update_config(
                user_id,
                conversation_id=None,
                config_id=config_id,
                config_name=data.name,
                settings=data.config.get("settings"),
                style=data.config.get("style"),
                agent_description=data.config.get("agent_description"),
                end_goal=data.config.get("end_goal"),
                branch_initialisation=data.config.get("branch_initialisation"),
            )

        except Exception as e:
            logger.exception(f"Error in /save_config_user API during in-memory update")
            return JSONResponse(
                content={
                    "error": str(e),
                    "config": tree_manager.config.to_json(),
                    "frontend_config": user["frontend_config"].to_json(),
                    "warnings": warnings,
                }
            )

        # Save the frontend config to memory and local file system
        if data.frontend_config is not None:
            try:
                await user_manager.update_frontend_config(user_id, data.frontend_config)
                save_frontend_config_to_file(user_id, user["frontend_config"].to_json())

            except Exception as e:
                if "Could not connect to Weaviate" in str(e):
                    warnings.append(
                        "Could not connect to Weaviate. "
                        "Please check your WCD_URL and WCD_API_KEY are correct."
                    )
                    return JSONResponse(
                        content={
                            "error": "",
                            "config": tree_manager.config.to_json(),
                            "frontend_config": user["frontend_config"].to_json(),
                            "warnings": warnings,
                        }
                    )
                else:
                    logger.exception(
                        f"Error in /save_config_user API during frontend config update"
                    )
                    return JSONResponse(
                        content={
                            "error": str(e),
                            "config": tree_manager.config.to_json(),
                            "frontend_config": user["frontend_config"].to_json(),
                            "warnings": warnings,
                        }
                    )

    if not user["frontend_config"].config["save_configs_to_weaviate"]:
        return JSONResponse(
            content={
                "error": "",
                "config": tree_manager.config.to_json(),
                "frontend_config": user["frontend_config"].to_json(),
                "warnings": warnings,
            }
        )

    # Save the config to the weaviate database
    try:

        # Retrieve the user info
        user = await user_manager.get_user_local(user_id)
        settings: Settings = user["tree_manager"].settings
        settings_dict = settings.to_json()

        # extract style, agent_description, end_goal, branch_initialisation from user
        style = user["tree_manager"].config.style
        agent_description = user["tree_manager"].config.agent_description
        end_goal = user["tree_manager"].config.end_goal
        branch_initialisation = user["tree_manager"].config.branch_initialisation

        # Check if the user has a valid save location
        if (
            user["frontend_config"].save_location_wcd_url == ""
            or user["frontend_config"].save_location_wcd_api_key == ""
        ):
            warnings.append(
                "No valid destination for config save location found. "
                "Config has not been saved to Weaviate. "
                "Saving to Weaviate requires a Config Storage WCD_URL and WCD_API_KEY."
            )
        else:

            settings_dict = encrypt_api_keys(settings_dict)

            # Connect to the weaviate database
            async with user[
                "frontend_config"
            ].save_location_client_manager.connect_to_async_client() as client:

                uuid = generate_uuid5(config_id)

                # Create a collection if it doesn't exist
                if await client.collections.exists("ELYSIA_CONFIG__"):
                    collection = client.collections.get("ELYSIA_CONFIG__")
                else:
                    collection = await client.collections.create(
                        "ELYSIA_CONFIG__",
                        vectorizer_config=wc.Configure.Vectorizer.none(),
                        inverted_index_config=wc.Configure.inverted_index(
                            index_timestamps=True
                        ),
                        properties=[
                            Property(
                                name="name",
                                data_type=DataType.TEXT,
                            ),
                            Property(
                                name="style",
                                data_type=DataType.TEXT,
                            ),
                            Property(
                                name="agent_description",
                                data_type=DataType.TEXT,
                            ),
                            Property(
                                name="end_goal",
                                data_type=DataType.TEXT,
                            ),
                            Property(
                                name="branch_initialisation",
                                data_type=DataType.TEXT,
                            ),
                            Property(
                                name="user_id",
                                data_type=DataType.TEXT,
                            ),
                            Property(
                                name="config_id",
                                data_type=DataType.TEXT,
                            ),
                            Property(
                                name="default",
                                data_type=DataType.BOOL,
                            ),
                        ],
                    )

                format_dict_to_serialisable(settings_dict)

                if settings_dict == {}:
                    settings_dict["null"] = "null"

                if "API_KEYS" in settings_dict and settings_dict["API_KEYS"] == {}:
                    settings_dict["API_KEYS"]["null"] = "null"

                config_item = {
                    "name": data.name,
                    "settings": settings_dict,
                    "style": style,
                    "agent_description": agent_description,
                    "end_goal": end_goal,
                    "branch_initialisation": branch_initialisation,
                    "frontend_config": user["frontend_config"].config,
                    "config_id": config_id,
                    "user_id": user_id,
                    "default": data.default,
                }

                # if the config is a default config, set all other default configs to False
                if data.default:
                    existing_default_config = await collection.query.fetch_objects(
                        filters=Filter.all_of(
                            [
                                Filter.by_property("default").equal(True),
                                Filter.by_property("user_id").equal(user_id),
                            ]
                        )
                    )
                    for item in existing_default_config.objects:
                        await collection.data.update(
                            properties={"default": False}, uuid=item.uuid
                        )

                # save the config to the weaviate database
                if await collection.data.exists(uuid=uuid):
                    await collection.data.update(properties=config_item, uuid=uuid)
                else:
                    await collection.data.insert(config_item, uuid=uuid)

    except Exception as e:
        logger.exception(f"Error in /save_config_user API during weaviate save")
        return JSONResponse(
            content={
                "error": str(e),
                "config": tree_manager.config.to_json(),
                "frontend_config": user["frontend_config"].to_json(),
                "warnings": warnings,
            }
        )

    logger.debug(f"Backend config returned: {tree_manager.config.to_json()}")
    logger.debug(f"Frontend config returned: {user['frontend_config'].to_json()}")

    return JSONResponse(
        content={
            "error": "",
            "config": tree_manager.config.to_json(),
            "frontend_config": user["frontend_config"].to_json(),
            "warnings": warnings,
        }
    )


@router.post("/{user_id}/{config_id}/load")
async def load_config_user(
    user_id: str,
    config_id: str,
    user_manager: UserManager = Depends(get_user_manager),
):
    """
    Load a config from the weaviate database for a user.
    This will set the default config for all new trees associated with the user.
    It will also propagate the config to any trees set with default settings.
    It will not affect trees that have been set up with their own config.

    It will also update the ClientManager to use the new API keys.

    Args:
        user_id (str): The user ID.
        config_id (str): The config ID you want to load.
        user_manager (UserManager): The user manager instance.

    Returns:
        JSONResponse: A JSON response with the following attributes:
            error (str): An error message. Empty if no error.
            config (dict): A dictionary of the updated config values.
    """
    logger.debug(f"/load_config_user API request received")
    logger.debug(f"User ID: {user_id}")
    logger.debug(f"Config ID: {config_id}")

    warnings = []

    # Retrieve the user info
    try:
        user = await user_manager.get_user_local(user_id)
        tree_manager: TreeManager = user["tree_manager"]
        frontend_config: FrontendConfig = user["frontend_config"]
    except Exception as e:
        logger.exception(f"Error in /load_config_user API during user retrieval")
        return JSONResponse(
            content={
                "error": str(e),
                "config": {},
                "frontend_config": {},
                "warnings": warnings,
            }
        )

    try:

        # check if the user has a valid save location
        if (
            frontend_config.save_location_wcd_url == ""
            or frontend_config.save_location_wcd_api_key == ""
        ):
            raise Exception("WCD URL or API key not found.")

        if (
            frontend_config.save_location_wcd_url == ""
            or frontend_config.save_location_wcd_api_key == ""
        ):
            raise Exception(
                "No valid destination for config load location found. "
                "Please update the save location using the /update_save_location API."
            )

        # Retrieve the config from the weaviate database
        async with (
            frontend_config.save_location_client_manager.connect_to_async_client() as client
        ):
            uuid = generate_uuid5(config_id)
            collection = client.collections.get("ELYSIA_CONFIG__")
            config_response = await collection.query.fetch_object_by_id(uuid=uuid)
            if config_response is None:
                raise Exception(f"Config with ID {config_id} not found.")

            config_item: dict = config_response.properties  # type: ignore

        format_dict_to_serialisable(config_item)

        if "settings" in config_item:
            try:
                config_item["settings"] = decrypt_api_keys(config_item["settings"])
            except InvalidToken:
                warnings.append(
                    "Invalid token for config. You will need to re-enter your API keys in the settings."
                )
                config_item["settings"]["API_KEYS"] = {}
                config_item["settings"]["WCD_API_KEY"] = ""

        # Rename the keys to the correct format
        renamed_config = rename_keys(config_item)

        # Load the settings
        await user_manager.update_config(
            user_id,
            conversation_id=None,
            config_id=renamed_config["config_id"],
            config_name=renamed_config["name"],
            settings=renamed_config["settings"],
            style=renamed_config["style"],
            agent_description=renamed_config["agent_description"],
            end_goal=renamed_config["end_goal"],
            branch_initialisation=renamed_config["branch_initialisation"],
        )

        # Update the frontend config
        await user_manager.update_frontend_config(
            user_id, renamed_config["frontend_config"]
        )

    except Exception as e:
        tree_manager = user["tree_manager"]
        config = tree_manager.config
        frontend_config = user["frontend_config"]
        logger.exception(f"Error in /load_config API")
        return JSONResponse(
            content={
                "error": str(e),
                "config": config.to_json(),
                "frontend_config": frontend_config.to_json(),
                "warnings": warnings,
            }
        )

    tree_manager = user["tree_manager"]
    config = tree_manager.config
    frontend_config = user["frontend_config"]
    return JSONResponse(
        content={
            "error": "",
            "config": config.to_json(),
            "frontend_config": frontend_config.to_json(),
            "warnings": warnings,
        }
    )


@router.delete("/{user_id}/{config_id}")
async def delete_config(
    user_id: str,
    config_id: str,
    user_manager: UserManager = Depends(get_user_manager),
):
    """
    Delete a config from the weaviate database.
    """
    logger.debug(f"/delete_config API request received")
    logger.debug(f"User ID: {user_id}")
    logger.debug(f"Config ID: {config_id}")

    try:
        user = await user_manager.get_user_local(user_id)
        frontend_config: FrontendConfig = user["frontend_config"]
        async with (
            frontend_config.save_location_client_manager.connect_to_async_client()
        ) as client:
            collection = client.collections.get("ELYSIA_CONFIG__")
            uuid = generate_uuid5(config_id)
            await collection.data.delete_by_id(uuid)

    except Exception as e:
        logger.exception(f"Error in /delete_config API")
        return JSONResponse(content={"error": str(e)})

    return JSONResponse(content={"error": ""})


@router.get("/{user_id}/list")
async def list_configs(
    user_id: str,
    user_manager: UserManager = Depends(get_user_manager),
):
    """
    List all the configs for a user.

    Args:
        data (ListConfigsData): A class with the following attributes:
            user_id (str): The user ID.
        user_manager (UserManager): The user manager instance.

    Returns:
        JSONResponse: A JSON response with the following attributes:
            error (str): An error message. Empty if no error.
            configs (list): A list of config IDs.
    """
    logger.debug(f"/list_configs API request received")
    logger.debug(f"User ID: {user_id}")

    headers = {"Cache-Control": "no-cache"}
    warnings = []

    try:
        user = await user_manager.get_user_local(user_id)

        if (
            "frontend_config" not in user
            or "save_location_wcd_url" not in user["frontend_config"].__dict__
            or "save_location_wcd_api_key" not in user["frontend_config"].__dict__
            or user["frontend_config"].save_location_wcd_url is None
            or user["frontend_config"].save_location_wcd_api_key is None
            or user["frontend_config"].save_location_wcd_url == ""
            or user["frontend_config"].save_location_wcd_api_key == ""
        ):
            logger.warning(
                "In /list_configs API, "
                "no valid destination for config location found. "
                "Returning no error but an empty list of configs."
            )
            warnings.append(
                "No valid destination for config location found. "
                "Cannot show saved configs."
            )
            return JSONResponse(
                content={"error": "", "configs": [], "warnings": warnings},
                headers=headers,
            )

        if not user["frontend_config"].save_location_client_manager.is_client:
            logger.warning(
                "In /list_configs API, "
                "Client manager is not connected to a client. "
                "Returning no error but an empty list of configs."
            )
            warnings.append(
                "Weaviate client is not connected. Cannot show saved configs."
            )
            return JSONResponse(
                content={"error": "", "configs": [], "warnings": warnings},
                headers=headers,
            )

        if user_id is not None:
            user_id_filter = Filter.by_property("user_id").equal(user_id)
        else:
            user_id_filter = None

        async with user[
            "frontend_config"
        ].save_location_client_manager.connect_to_async_client() as client:

            if await client.collections.exists("ELYSIA_CONFIG__"):
                collection = client.collections.get("ELYSIA_CONFIG__")
                len_collection = (
                    await collection.aggregate.over_all(total_count=True)
                ).total_count

                response = await collection.query.fetch_objects(
                    limit=len_collection,
                    sort=Sort.by_update_time(ascending=False),
                    return_metadata=MetadataQuery(last_update_time=True),
                    filters=user_id_filter,
                )

                configs = [
                    {
                        "config_id": obj.properties["config_id"],
                        "name": obj.properties["name"],
                        "default": obj.properties["default"],
                        "last_update_time": format_datetime(
                            obj.metadata.last_update_time
                        ),
                    }
                    for obj in response.objects
                ]
            else:
                configs = []

    except Exception as e:
        logger.exception(f"Error in /list_configs API")
        return JSONResponse(
            content={"error": str(e), "configs": [], "warnings": warnings},
            headers=headers,
        )

    return JSONResponse(
        content={"error": "", "configs": configs, "warnings": warnings}, headers=headers
    )



import logging
from rich.logging import RichHandler
import litellm
import dspy

# Initialize the logger as a module-level variable
_logger = None


def get_logger():
    """Get the singleton logger instance."""
    global _logger
    if _logger is None:
        # Configure basic logging if not already configured
        if not logging.root.handlers:
            FORMAT = "%(message)s"
            logging.basicConfig(
                level="DEBUG",
                format=FORMAT,
                datefmt="[%X]",
                handlers=[RichHandler(rich_tracebacks=True, markup=True)],
            )

        # Set log levels for specific loggers
        logging.getLogger("httpx").setLevel(logging.WARNING)
        logging.getLogger("httpcore").setLevel(logging.WARNING)
        logging.getLogger("urllib3").setLevel(logging.WARNING)
        logging.getLogger("asyncio").setLevel(logging.INFO)
        logging.getLogger("fastapi").setLevel(logging.INFO)
        logging.getLogger("uvicorn").setLevel(logging.INFO)
        logging.getLogger("starlette").setLevel(logging.INFO)
        logging.getLogger("grpc").setLevel(logging.INFO)
        logging.getLogger("grpc._channel").setLevel(logging.WARNING)
        logging.getLogger("dspy").setLevel(logging.WARNING)
        logging.getLogger("dspy").propagate = False
        logging.getLogger("LiteLLM").setLevel(logging.WARNING)
        logging.getLogger("LiteLLM").propagate = False
        logging.getLogger("LiteLLM Router").setLevel(logging.WARNING)
        logging.getLogger("LiteLLM Router").propagate = False
        logging.getLogger("LiteLLM Proxy").setLevel(logging.WARNING)
        logging.getLogger("LiteLLM Proxy").propagate = False
        logging.getLogger("matplotlib").setLevel(logging.WARNING)
        logging.getLogger("matplotlib").propagate = False

        _logger = logging.getLogger("rich")
        _logger.setLevel(logging.DEBUG)

    return _logger


def set_log_level(level: str):
    """Set the level for the main application logger."""
    logger = get_logger()
    logger.setLevel(level)


# Initialize the logger
logger = get_logger()

dspy.disable_litellm_logging()
dspy.disable_logging()
litellm.suppress_debug_info = True



from elysia.api.services.user import UserManager

# Create the singleton instance at module level
user_manager = UserManager()


def get_user_manager() -> UserManager:
    """Get the singleton UserManager instance."""
    return user_manager



from elysia.__metadata__ import (
    __version__,
    __name__,
    __description__,
    __url__,
    __author__,
    __author_email__,
)

from elysia.tree.tree import Tree
from elysia.objects import (
    Tool,
    Return,
    Text,
    Response,
    Update,
    Status,
    Warning,
    Error,
    Completed,
    Result,
    Retrieval,
    tool,
)
from elysia.preprocessing.collection import (
    preprocess,
    preprocessed_collection_exists,
    edit_preprocessed_collection,
    delete_preprocessed_collection,
    view_preprocessed_collection,
)
from elysia.config import Settings, settings, configure, smart_setup, set_from_env



import os
from cryptography.fernet import Fernet
from dotenv import set_key, load_dotenv
from copy import deepcopy


def encrypt_api_keys(settings_dict: dict):

    load_dotenv()

    settings_dict_copy = deepcopy(settings_dict)

    # Check if an auth key is set in the environment variables
    if "FERNET_KEY" in os.environ:
        auth_key = os.environ["FERNET_KEY"]
    else:  # create one
        auth_key = Fernet.generate_key().decode("utf-8")
        set_key(".env", "FERNET_KEY", auth_key)

    # Encode all api keys
    f = Fernet(auth_key.encode("utf-8"))
    if "API_KEYS" in settings_dict_copy:
        for key in settings_dict_copy["API_KEYS"]:
            if key != "null":
                settings_dict_copy["API_KEYS"][key] = f.encrypt(
                    settings_dict_copy["API_KEYS"][key].encode("utf-8")
                ).decode("utf-8")

    if "WCD_API_KEY" in settings_dict_copy:
        settings_dict_copy["WCD_API_KEY"] = f.encrypt(
            settings_dict_copy["WCD_API_KEY"].encode("utf-8")
        ).decode("utf-8")

    return settings_dict_copy


def decrypt_api_keys(settings_dict: dict):

    load_dotenv()

    settings_dict_copy = deepcopy(settings_dict)

    if "FERNET_KEY" in os.environ:
        auth_key = os.environ["FERNET_KEY"]
    else:  # create one
        auth_key = Fernet.generate_key().decode("utf-8")
        set_key(".env", "FERNET_KEY", auth_key)

    # decode all api keys
    f = Fernet(auth_key.encode("utf-8"))

    if "API_KEYS" in settings_dict_copy:
        for key in settings_dict_copy["API_KEYS"]:
            if key != "null":
                settings_dict_copy["API_KEYS"][key] = f.decrypt(
                    settings_dict_copy["API_KEYS"][key].encode("utf-8")
                ).decode("utf-8")

    if "WCD_API_KEY" in settings_dict_copy:
        settings_dict_copy["WCD_API_KEY"] = f.decrypt(
            settings_dict_copy["WCD_API_KEY"].encode("utf-8")
        ).decode("utf-8")

    return settings_dict_copy



import os
from fastapi.responses import JSONResponse
from pathlib import Path
from logging import Logger
import json
import random
import pytest
from weaviate.util import generate_uuid5
from elysia.api.api_types import (
    InitialiseTreeData,
    SaveConfigUserData,
    SaveConfigTreeData,
    UpdateFrontendConfigData,
)
from elysia.api.dependencies.common import get_user_manager
from elysia.api.routes.init import initialise_tree, initialise_user
from elysia.api.routes.user_config import (
    load_config_user,
    delete_config,
    save_config_user,
    new_user_config,
    list_configs,
    update_frontend_config,
    get_models,
)
from elysia.api.routes.tree_config import (
    load_config_tree,
    change_config_tree,
    new_tree_config,
)
from elysia.tree.tree import Tree
from elysia.api.core.log import logger, set_log_level
from elysia.api.services.tree import TreeManager
from elysia.api.services.user import UserManager
from elysia.api.utils.config import FrontendConfig

set_log_level("CRITICAL")


def read_response(response: JSONResponse):
    return json.loads(response.body)


def get_frontend_config_file_path(user_id: str) -> Path:
    elysia_package_dir = Path(__file__).parent.parent.parent.parent  # Gets to elysia/
    config_dir = elysia_package_dir / "elysia" / "api" / "user_configs"
    config_file = config_dir / f"frontend_config_{user_id}.json"
    return config_file


async def load_frontend_config_from_file(
    user_id: str, logger: Logger
) -> FrontendConfig:
    try:
        elysia_package_dir = Path(
            __file__
        ).parent.parent.parent.parent  # Gets to elysia/
        config_dir = elysia_package_dir / "elysia" / "api" / "user_configs"
        config_file = config_dir / f"frontend_config_{user_id}.json"

        if not config_file.exists():
            return FrontendConfig(logger=logger)

        with open(config_file, "r") as f:
            fe_config = await FrontendConfig.from_json(json.load(f), logger)
            return fe_config

    except Exception as e:
        logger.error(f"Error in save_frontend_config_to_file: {str(e)}")
        return FrontendConfig(logger=logger)


async def initialise_user_and_tree(user_id: str, conversation_id: str):
    user_manager = get_user_manager()

    response = await initialise_user(
        user_id,
        user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    response = await initialise_tree(
        user_id,
        conversation_id,
        InitialiseTreeData(
            low_memory=False,
        ),
        user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""


@pytest.mark.asyncio
async def test_get_models():
    """
    Test the endpoint for getting the models.
    """
    user_manager = get_user_manager()
    response = await get_models()
    response = read_response(response)
    assert response["error"] == ""
    assert "models" in response
    assert len(response["models"]) > 0
    assert "openai" in response["models"]
    assert "anthropic" in response["models"]
    assert "gemini" in response["models"]


@pytest.mark.asyncio
async def test_config_workflow():
    """
    Test the config workflow.
    """
    user_manager = get_user_manager()
    user_id = "test_user_config_workflow"
    conversation_id_1 = "test_conversation_workflow_1"
    conversation_id_2 = "test_conversation_workflow_2"
    config_id = f"test_config_conversation_config_workflow"
    config_name = "Test config workflow"

    await initialise_user_and_tree(user_id, conversation_id_1)

    # Test changing config locally
    user_settings = {
        "BASE_MODEL": "gpt-4o-mini",
        "COMPLEX_MODEL": "gpt-4o",
        "BASE_PROVIDER": "openai",
        "COMPLEX_PROVIDER": "openai",
    }
    response = await save_config_user(
        user_id=user_id,
        config_id=config_id,
        data=SaveConfigUserData(
            name=config_name,
            default=False,
            config={
                "settings": user_settings,
                "style": "Professional style for user",
            },
            frontend_config={
                "save_configs_to_weaviate": False,
            },
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # Check tree manager settings have NOT changed because default=False
    tree_manager = user_manager.users[user_id]["tree_manager"]
    assert tree_manager.settings.BASE_MODEL != "gpt-4o-mini"
    assert tree_manager.settings.COMPLEX_MODEL != "gpt-4o"
    assert tree_manager.settings.BASE_PROVIDER != "openai"
    assert tree_manager.settings.COMPLEX_PROVIDER != "openai"

    # save again with default=True
    response = await save_config_user(
        user_id=user_id,
        config_id=config_id,
        data=SaveConfigUserData(
            name=config_name,
            default=True,
            config={
                "settings": user_settings,
                "style": "Professional style for user",
            },
            frontend_config={
                "save_configs_to_weaviate": False,
            },
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    assert tree_manager.settings.BASE_MODEL == "gpt-4o-mini"
    assert tree_manager.settings.COMPLEX_MODEL == "gpt-4o"
    assert tree_manager.settings.BASE_PROVIDER == "openai"
    assert tree_manager.settings.COMPLEX_PROVIDER == "openai"

    # Verify tree 1 inherits user model settings including style
    tree1: Tree = await user_manager.get_tree(user_id, conversation_id_1)
    assert tree1.settings.BASE_MODEL == "gpt-4o-mini"
    assert tree1.settings.COMPLEX_MODEL == "gpt-4o"
    assert tree1.tree_data.atlas.style == "Professional style for user"

    # Create tree 2
    response = await initialise_tree(
        user_id,
        conversation_id_2,
        data=InitialiseTreeData(
            low_memory=True,
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # Tree 2 should inherit the new settings
    tree2: Tree = await user_manager.get_tree(user_id, conversation_id_2)
    assert tree2.settings.BASE_MODEL == "gpt-4o-mini"
    assert tree2.settings.COMPLEX_MODEL == "gpt-4o"

    # Change tree 1 settings
    tree1_settings = {
        "BASE_MODEL": "claude-3-haiku",
        "COMPLEX_MODEL": "claude-3-opus",
        "BASE_PROVIDER": "anthropic",
        "COMPLEX_PROVIDER": "anthropic",
    }
    response = await change_config_tree(
        user_id=user_id,
        conversation_id=conversation_id_1,
        data=SaveConfigTreeData(
            settings=tree1_settings,
            style="Custom tree 1 style",
            agent_description="Custom tree 1 agent description",
            end_goal="Custom tree 1 end goal",
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # Verify tree 1 has its own settings now
    tree1 = await user_manager.get_tree(user_id, conversation_id_1)
    assert tree1.settings.BASE_MODEL == "claude-3-haiku"
    assert tree1.settings.COMPLEX_MODEL == "claude-3-opus"
    assert tree1.tree_data.atlas.style == "Custom tree 1 style"

    # change user settings again
    new_user_settings = {
        "BASE_MODEL": "gemini-pro",
        "COMPLEX_MODEL": "gemini-ultra",
        "BASE_PROVIDER": "google",
        "COMPLEX_PROVIDER": "google",
    }
    response = await save_config_user(
        user_id=user_id,
        config_id=config_id,
        data=SaveConfigUserData(
            name=config_name,
            default=True,
            config={
                "settings": new_user_settings,
            },
            frontend_config={
                "save_configs_to_weaviate": False,
            },
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # Tree 1 should keep its own settings
    tree1 = await user_manager.get_tree(user_id, conversation_id_1)
    assert tree1.settings.BASE_MODEL == "claude-3-haiku"
    assert tree1.settings.COMPLEX_MODEL == "claude-3-opus"

    # But tree 2 should inherit updated user settings
    tree2 = await user_manager.get_tree(user_id, conversation_id_2)
    assert tree2.settings.BASE_MODEL == "gemini-pro"
    assert tree2.settings.COMPLEX_MODEL == "gemini-ultra"


@pytest.mark.asyncio
async def test_change_config_tree():
    """
    Test the endpoint for changing the config of a tree.
    """
    user_manager = get_user_manager()
    user_id = "test_user_change_config_tree"
    conversation_id = "test_conversation_change_config_tree"

    await initialise_user_and_tree(user_id, conversation_id)

    # get initial settings
    tree: Tree = await user_manager.get_tree(user_id, conversation_id)
    initial_base_model = tree.settings.BASE_MODEL
    initial_style = tree.tree_data.atlas.style

    # change tree config
    new_settings = {
        "BASE_MODEL": "gpt-3.5-turbo",
        "COMPLEX_MODEL": "gpt-4-turbo",
        "BASE_PROVIDER": "openai",
        "COMPLEX_PROVIDER": "openai",
        # Try including API keys, which should be ignored
        "openai_api_key": "test_key_should_be_ignored",
        "wcd_url": "should_be_ignored_too",
    }
    new_style = "Technical and detailed."
    new_agent_description = "You are a technical assistant."
    new_end_goal = "Provide technical details to the user."
    new_branch_init = "empty"

    response = await change_config_tree(
        user_id=user_id,
        conversation_id=conversation_id,
        data=SaveConfigTreeData(
            settings=new_settings,
            style=new_style,
            agent_description=new_agent_description,
            end_goal=new_end_goal,
            branch_initialisation=new_branch_init,
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # Tree settings should have been updated
    tree = await user_manager.get_tree(user_id, conversation_id)
    assert tree.settings.BASE_MODEL == "gpt-3.5-turbo"
    assert tree.settings.COMPLEX_MODEL == "gpt-4-turbo"
    assert tree.settings.BASE_PROVIDER == "openai"
    assert tree.settings.COMPLEX_PROVIDER == "openai"
    assert tree.tree_data.atlas.style == new_style
    assert tree.tree_data.atlas.agent_description == new_agent_description
    assert tree.tree_data.atlas.end_goal == new_end_goal
    # assert tree.branch_initialisation == new_branch_init

    # But not api keys
    assert (
        "openai_api_key" not in tree.settings.to_json()
        or tree.settings.to_json()["openai_api_key"] != "test_key_should_be_ignored"
    )
    assert tree.settings.WCD_URL != "should_be_ignored_too"

    # user-level settings should remain unchanged
    user = await user_manager.get_user_local(user_id)
    tree_manager = user["tree_manager"]
    assert tree_manager.settings.BASE_MODEL == initial_base_model
    assert tree_manager.config.style == initial_style


@pytest.mark.asyncio
async def test_new_config():
    """
    Test the endpoint for creating a new config.
    """
    user_manager = get_user_manager()
    user_id = "test_user_new_config"
    conversation_id = "test_conversation_new_config"
    config_id = f"test_new_config"
    config_name = "Test new config"

    await initialise_user_and_tree(user_id, conversation_id)

    # create a new config
    response = await new_user_config(
        user_id=user_id,
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == "", response["error"]
    new_user_config_id = response["config"]["id"]

    # save config
    response = await save_config_user(
        user_id=user_id,
        config_id=new_user_config_id,
        data=SaveConfigUserData(
            name=config_name,
            id=new_user_config_id,
            default=True,
            config=response["config"],
            frontend_config={
                "save_configs_to_weaviate": False,
            },
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == "", response["error"]

    response = await initialise_user(
        user_id,
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == "", response["error"]
    assert response["config"]["id"] == new_user_config_id


@pytest.mark.asyncio
async def test_save_config_no_api_keys():
    """
    Test that a config can be saved with no api keys.
    """
    user_manager = get_user_manager()
    user_id = "test_user_save_config_no_api_keys"
    conversation_id = "test_conversation_save_config_no_api_keys"
    config_id = f"test_config_conversation_save_config_no_api_keys"
    config_name = "Test save config no api keys"

    await initialise_user_and_tree(user_id, conversation_id)

    # Get tree manager in advance
    user = await user_manager.get_user_local(user_id)
    tree_manager = user["tree_manager"]

    # Save a config with no api keys
    response = await save_config_user(
        user_id=user_id,
        config_id=config_id,
        data=SaveConfigUserData(
            name=config_name,
            default=True,
            config={
                "settings": {
                    "BASE_MODEL": "gpt-4o-mini",
                    "COMPLEX_MODEL": "gpt-4o",
                    "BASE_PROVIDER": "openai",
                    "COMPLEX_PROVIDER": "openai",
                    "API_KEYS": {},
                },
                "style": "Professional and concise.",
                "agent_description": "You are a helpful assistant that searches data.",
                "end_goal": "Provide a summary of the data.",
                "branch_initialisation": "one_branch",
            },
            frontend_config={
                "save_configs_to_weaviate": False,
            },
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""


@pytest.mark.asyncio
async def test_change_config_user():
    """
    Test changing a user config propagates to the TreeManager and the trees.
    """
    user_manager = get_user_manager()
    user_id = "test_user_change_config_user"
    conversation_id = "test_conversation_change_config_user"
    config_id = f"test_config_conversation_change_config_user"
    config_name = "Test change config user"

    await initialise_user_and_tree(user_id, conversation_id)

    # Get tree manager in advance
    user = await user_manager.get_user_local(user_id)
    tree_manager = user["tree_manager"]

    # Set frontend config to not save configs to weaviate
    response = await save_config_user(
        user_id=user_id,
        config_id=config_id,
        data=SaveConfigUserData(
            name=config_name,
            default=True,
            config={},
            frontend_config={
                "save_configs_to_weaviate": False,
            },
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == "", response["error"]
    assert response["warnings"] == [], "\n".join(response["warnings"])

    # Change user config
    new_settings = {
        "BASE_MODEL": "gpt-4o-mini",
        "COMPLEX_MODEL": "gpt-4o",
        "BASE_PROVIDER": "openai",
        "COMPLEX_PROVIDER": "openai",
    }
    new_style = "Professional and concise."
    new_agent_description = "You are a helpful assistant that searches data."
    new_branch_init = "one_branch"
    old_end_goal = tree_manager.config.end_goal

    response = await save_config_user(
        user_id=user_id,
        config_id=config_id,
        data=SaveConfigUserData(
            name=config_name,
            default=True,
            config={
                "settings": new_settings,
                "style": new_style,
                "agent_description": new_agent_description,
                "branch_initialisation": new_branch_init,
            },
            frontend_config={
                "save_configs_to_weaviate": False,
            },
        ),
        user_manager=user_manager,
    )

    # Verify response is OK
    response = read_response(response)
    assert response["error"] == "", response["error"]
    assert response["warnings"] == [], "\n".join(response["warnings"])

    assert tree_manager.settings.BASE_MODEL == "gpt-4o-mini"
    assert tree_manager.settings.COMPLEX_MODEL == "gpt-4o"
    assert tree_manager.settings.BASE_PROVIDER == "openai"
    assert tree_manager.settings.COMPLEX_PROVIDER == "openai"

    assert tree_manager.config.id == config_id
    assert tree_manager.config.style == new_style
    assert tree_manager.config.agent_description == new_agent_description
    assert (
        tree_manager.config.end_goal == old_end_goal
    )  # this parameter wasn't set so it should be the same as before
    assert tree_manager.config.branch_initialisation == new_branch_init

    # Verify model settings propagated to existing tree
    tree: Tree = await user_manager.get_tree(user_id, conversation_id)
    assert tree.settings.BASE_MODEL == "gpt-4o-mini"
    assert tree.settings.COMPLEX_MODEL == "gpt-4o"
    assert tree.settings.BASE_PROVIDER == "openai"
    assert tree.settings.COMPLEX_PROVIDER == "openai"

    # And style, agent_description, etc. should have propagated to existing tree
    assert tree.tree_data.atlas.style == new_style
    assert tree.tree_data.atlas.agent_description == new_agent_description
    assert tree.branch_initialisation == new_branch_init


@pytest.mark.asyncio
@pytest.mark.skip(reason="Skipping frontend config local test")
async def test_frontend_config_local():
    """
    Test that the frontend config is saved to a local file and loads correctly.
    """
    user_manager = get_user_manager()
    user_id = "test_user_frontend_config_local"
    conversation_id = "test_conversation_frontend_config_local"
    config_id = f"test_frontend_config_local"
    config_name = "Test frontend config local"

    await initialise_user_and_tree(user_id, conversation_id)

    # create a config and save
    user = await user_manager.get_user_local(user_id)
    tree_manager: TreeManager = user["tree_manager"]

    # change config
    response = await save_config_user(
        user_id=user_id,
        config_id=config_id,
        data=SaveConfigUserData(
            name=config_name,
            default=True,
            config={},
            frontend_config={
                "save_trees_to_weaviate": False,
                "save_configs_to_weaviate": False,
            },
        ),
        user_manager=user_manager,
    )
    response = read_response(response)

    # check existence of local file
    # assert os.path.exists(get_frontend_config_file_path(user_id))
    # skip for now

    # check the local file has the correct values
    fe_config = await load_frontend_config_from_file(user_id, logger)
    fe_config = fe_config.to_json()

    assert fe_config["save_trees_to_weaviate"] == False
    assert fe_config["save_configs_to_weaviate"] == False

    # remove user from user manager
    del user_manager.users[user_id]

    # add user back to user manager
    await user_manager.add_user_local(user_id)

    # check the frontend config is loaded from the local file and correct settings
    user = await user_manager.get_user_local(user_id)
    assert user["frontend_config"].config["save_trees_to_weaviate"] == False
    assert user["frontend_config"].config["save_configs_to_weaviate"] == False


@pytest.mark.asyncio
async def test_new_config_no_overwrite():
    """
    Test creating a new config does not overwrite existing config.
    """
    user_manager = get_user_manager()
    user_id = "test_user_new_config_no_overwrite"
    conversation_id = "test_conversation_new_config_no_overwrite"
    config_name = "Test new config no overwrite"

    await initialise_user_and_tree(user_id, conversation_id)

    # get config id
    user = await user_manager.get_user_local(user_id)
    config_id = user["tree_manager"].config.id

    # save some values
    response = await save_config_user(
        user_id=user_id,
        config_id=config_id,
        data=SaveConfigUserData(
            name=config_name,
            id=config_id,
            default=True,
            config={
                "settings": {
                    "BASE_MODEL": "gpt-4o-mini",
                    "BASE_PROVIDER": "openai",
                    "COMPLEX_MODEL": "gpt-4o",
                    "COMPLEX_PROVIDER": "openai",
                    "API_KEYS": {
                        "openai_api_key": "some-random-openai-api-key",
                    },
                },
                "style": "New style for no overwrite",
                "agent_description": "New agent description for no overwrite",
                "end_goal": "New end goal for no overwrite",
                "branch_initialisation": "one_branch",
            },
            frontend_config={
                "save_trees_to_weaviate": False,
                "save_configs_to_weaviate": False,
            },
        ),
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""

    # create a new config
    response = await new_user_config(
        user_id=user_id,
        user_manager=user_manager,
    )
    response = read_response(response)
    assert response["error"] == ""
    new_config_id = response["config"]["id"]

    # check the user values were NOT saved
    user = await user_manager.get_user_local(user_id)
    assert user["tree_manager"].config.id == config_id
    assert user["tree_manager"].config.settings.BASE_MODEL == "gpt-4o-mini"
    assert user["tree_manager"].config.settings.BASE_PROVIDER == "openai"
    assert user["tree_manager"].config.settings.COMPLEX_MODEL == "gpt-4o"
    assert user["tree_manager"].config.settings.COMPLEX_PROVIDER == "openai"
    assert "openai_api_key" in user["tree_manager"].config.settings.API_KEYS
    assert (
        user["tree_manager"].config.settings.API_KEYS["openai_api_key"]
        == "some-random-openai-api-key"
    )



from pydantic import BaseModel
from pydantic.fields import Field

from typing import Optional, Literal, Any
from datetime import datetime
from elysia.objects import Result


class BarChartData(BaseModel):
    x_labels: list[str | int | float] = Field(
        description="The shared labels for the x-axis. Must be all string or all numeric. Cannot mix.",
        min_length=1,
        max_length=10,
    )
    y_values: dict[str, list[int | float]] = Field(
        description="A dictionary of data points for the y-axis, with the key being the category/group for the data points.",
        min_length=1,
        max_length=10,
    )


class BarChart(BaseModel):
    title: str
    description: str
    x_axis_label: str
    y_axis_label: str
    data: BarChartData = Field(
        description="A dictionary of bar chart data points, with the key being the category/group for the data points."
    )


class HistogramData(BaseModel):
    distribution: list[float | int]


class HistogramChart(BaseModel):
    title: str
    description: str
    data: dict[str, HistogramData] = Field(
        description="A dictionary of histogram data points, with the key being the category/grouping for the data points."
    )


class ScatterOrLineDataPoint(BaseModel):
    value: int | float | datetime | str
    label: Optional[str] = Field(
        default="",
        description="If a data point is highlighted, you can label it.",
        min_length=1,
        max_length=50,
    )


class ScatterOrLineYAxisData(BaseModel):
    label: str = Field(
        description="The label for the data points going into the y-axis. "
    )
    kind: Literal["scatter", "line"] = Field(
        description="Whether the data points (for this y-axis data) are scatter points, or a line plot. "
    )
    data_points: list[ScatterOrLineDataPoint]


class ScatterOrLineDataPoints(BaseModel):
    x_axis: list[ScatterOrLineDataPoint] = Field(
        default=[],
        description=(
            "The shared x-axis data points. This MUST be populated. "
            "If you think there is no relevant data for the x-axis, "
            "do not leave it empty. X-values are required for Y-values to be plotted. "
        ),
    )
    y_axis: list[ScatterOrLineYAxisData] = Field(
        default=[],
        description=(
            "Each element of the list has a label for the data points going into the y-axis. "
            "Each one is either a set of scatter points, or a line plot. "
            "Which can be combined in a single chart, and will be automatically separated into groups in the same chart (with labels). "
            "There can be multiple sets of data, under different labels, indicating different groups of data. "
            "These will be plotted on the same y-axis, but with different colours, shapes, etc."
        ),
        min_length=1,
        max_length=20,
    )
    normalize_y_axis: Optional[bool] = Field(
        default=False,
        description=(
            "If true, all values in the y-axis will be normalized to the range 0-1. "
            "Useful if the two values are not on the same scale. "
            "Do not modify the labels, or the values, these will be automatically changed."
        ),
    )


class ScatterOrLineChart(BaseModel):
    title: str
    description: str
    x_axis_label: str
    y_axis_label: str
    data: ScatterOrLineDataPoints = Field(
        description="A dictionary of scatter or line data points, with the key being the label for the data points."
    )


class ChartResult(Result):
    def __init__(
        self,
        charts: list[BarChart | HistogramChart | ScatterOrLineChart],
        chart_type: Literal["bar", "histogram", "scatter_or_line"],
        title: str = "",
        metadata: dict[str, Any] = {},
    ):
        if len(charts) == 1 and title == "":
            title = charts[0].title

        super().__init__(
            objects=[chart.model_dump() for chart in charts],
            metadata={"chart_title": title, "chart_type": chart_type, **metadata},
            payload_type=f"{chart_type}_chart",
            mapping=None,
            name=f"{chart_type}_chart",
        )

    def llm_parse(self):
        if "impossible" in self.metadata:
            out = f"Model judged creation of {self.metadata['chart_type']} chart to be impossible for reason: "
            if "impossible_reasoning" in self.metadata:
                out += f"'{self.metadata['impossible_reasoning']}'. "
            out += "Returning to the decision tree..."
            return out
        else:
            return f"Created {len(self.objects)} {self.metadata['chart_type']} chart(s) titled '{self.metadata['chart_title']}'."
